"""Data objects about Multi-Signature."""

from __future__ import annotations

from typing import List

import attr

from terra_sdk.core.compact_bit_array import CompactBitArray
from terra_sdk.core.public_key import LegacyAminoMultisigPublicKey, SimplePublicKey
from terra_sdk.core.signature_v2 import Descriptor
from terra_sdk.core.signature_v2 import Multi as MultiDescriptor
from terra_sdk.core.signature_v2 import SignatureV2
from terra_sdk.util.json import JSONSerializable

__all__ = ["MultiSignature"]


@attr.s
class MultiSignature(JSONSerializable):
    bitarray: CompactBitArray = attr.ib(init=False)
    signatures: List[Descriptor] = attr.ib(init=False)
    multisig_pubkey: LegacyAminoMultisigPublicKey = attr.ib()

    def __attrs_post_init__(self):
        n = len(self.multisig_pubkey.public_keys)
        self.bitarray = CompactBitArray.from_bits(n)
        self.signatures = []

    def append_signature(self, signature_data: Descriptor, index: int):
        new_idx = self.bitarray.num_true_bits_before(index)

        # in case of signature already exists, just replace
        if self.bitarray.get_index(index):
            self.signatures[new_idx] = signature_data
            return

        self.bitarray.set_index(index, True)
        # optimization
        if new_idx == len(self.signatures):
            self.signatures.append(signature_data)
            return

        self.signatures.insert(new_idx, signature_data)

    def append_signature_from_pubkey(
        self, signature_data: Descriptor, public_key: SimplePublicKey
    ):
        index = 0
        for i, v in enumerate(self.multisig_pubkey.public_keys):
            if v.key == public_key.key:
                index = i
                break
        if index < 0:
            raise ValueError("provided key doesn't exist in public_keys")
        self.append_signature(signature_data, index)

    def append_signature_v2s(self, signatures: List[SignatureV2]):
        for sig in signatures:
            if not isinstance(sig.public_key, SimplePublicKey):
                raise ValueError("non-SimplePublicKey cannot be used to sign multisig")
            self.append_signature_from_pubkey(sig.data, sig.public_key)

    def to_signature_descriptor(self) -> Descriptor:
        return Descriptor(
            multi=MultiDescriptor(bitarray=self.bitarray, signatures=self.signatures)
        )



"""feegrant module message types."""

from __future__ import annotations

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.feegrant.v1beta1 import (
    MsgGrantAllowance as MsgGrantAllowance_pb,
)
from terra_proto.cosmos.feegrant.v1beta1 import (
    MsgRevokeAllowance as MsgRevokeAllowance_pb,
)

from terra_sdk.core import AccAddress
from terra_sdk.core.msg import Msg

from .data import Allowance

__all__ = ["MsgGrantAllowance", "MsgRevokeAllowance"]


@attr.s
class MsgGrantAllowance(Msg):
    """
    MsgGrantAllowance adds permission for Grantee to spend up to Allowance
    of fees from the account of Granter.
    """

    granter: AccAddress = attr.ib()
    grantee: AccAddress = attr.ib()
    allowance: Allowance = attr.ib()

    type_amino = "cosmos-sdk/MsgGrantAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.MsgGrantAllowance"
    """"""
    prototype = MsgGrantAllowance_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "granter": self.granter,
                "grantee": self.grantee,
                "allowance": self.allowance.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "granter": self.granter,
            "grantee": self.grantee,
            "allowance": self.allowance.to_data(),
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgGrantAllowance:
        return cls(
            granter=data["granter"],
            grantee=data["grantee"],
            allowance=Allowance.from_data(data["allowance"]),
        )

    def to_proto(self) -> MsgGrantAllowance_pb:
        return MsgGrantAllowance_pb(
            granter=self.granter,
            grantee=self.grantee,
            allowance=Allowance.pack_any(self.allowance),
        )

    @classmethod
    def from_proto(cls, proto: MsgGrantAllowance_pb) -> MsgGrantAllowance:
        return cls(
            granter=proto.granter,
            grantee=proto.grantee,
            allowance=Allowance.unpack_any(proto.allowance),
        )


@attr.s
class MsgRevokeAllowance(Msg):
    """MsgRevokeAllowance remove permission any existing Allowance from Granter to Grantee."""

    granter: AccAddress = attr.ib()
    grantee: AccAddress = attr.ib()

    type_amino = "cosmos-sdk/MsgRevokeAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.MsgRevokeAllowance"
    """"""
    prototype = MsgRevokeAllowance_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"granter": self.granter, "grantee": self.grantee},
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgRevokeAllowance:
        return cls(granter=data["granter"], grantee=data["grantee"])

    def to_proto(self) -> MsgRevokeAllowance_pb:
        return MsgRevokeAllowance_pb(granter=self.granter, grantee=self.grantee)

    @classmethod
    def from_proto(cls, proto: MsgRevokeAllowance_pb) -> MsgRevokeAllowance:
        return cls(granter=proto.granter, grantee=proto.grantee)



"""Authz module data types."""

from __future__ import annotations

from datetime import datetime
from typing import List, Optional

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from dateutil import parser
from terra_proto.cosmos.authz.v1beta1 import (
    GenericAuthorization as GenericAuthorization_pb,
)
from terra_proto.cosmos.authz.v1beta1 import Grant as Grant_pb
from terra_proto.cosmos.bank.v1beta1 import SendAuthorization as SendAuthorization_pb
from terra_proto.cosmos.staking.v1beta1 import AuthorizationType
from terra_proto.cosmos.staking.v1beta1 import (
    StakeAuthorization as StakeAuthorization_pb,
)
from terra_proto.cosmos.staking.v1beta1 import (
    StakeAuthorizationValidators as StakeAuthorizationValidators_pb,
)

from terra_sdk.core import AccAddress, Coin, Coins
from terra_sdk.util.base import BaseTerraData
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "Authorization",
    "SendAuthorization",
    "GenericAuthorization",
    "StakeAuthorization",
    "AuthorizationGrant",
    "AuthorizationType",
]


class Authorization(BaseTerraData):
    """Base class for authorization types."""

    @staticmethod
    def from_amino(amino: dict) -> Authorization:
        from terra_sdk.util.parse_authorization import parse_authorization_amino

        return parse_authorization_amino(amino)

    @staticmethod
    def from_data(data: dict) -> Authorization:
        from terra_sdk.util.parse_authorization import parse_authorization

        return parse_authorization(data)

    @staticmethod
    def from_proto(proto: Any_pb) -> Authorization:
        from terra_sdk.util.parse_authorization import parse_authorization_proto

        return parse_authorization_proto(proto)

    @staticmethod
    def unpack_any(proto: Any_pb) -> Authorization:
        from terra_sdk.util.parse_authorization import parse_authorization_unpack_any

        return parse_authorization_unpack_any(proto)


@attr.s
class SendAuthorization(Authorization):
    """Type of :class:`Authorization` for :class:`MsgSend<terra_sdk.core.bank.msgs.MsgSend>`,
    which can be parameterized by setting a ``spend_limit`` allowance for the grantee.

    Args:
        spend_limit (Coins.Input): coins representing allowance of grant
    """

    type_amino = "cosmos-sdk/SendAuthorization"
    """"""
    type_url = "/cosmos.bank.v1beta1.SendAuthorization"
    """"""
    prototype = SendAuthorization_pb
    """"""

    spend_limit: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"spend_limit": self.spend_limit.to_amino()},
        }

    def to_data(self) -> dict:
        return {"@type": self.type_url, "spend_limit": self.spend_limit.to_data()}

    @classmethod
    def from_data(cls, data: dict) -> SendAuthorization:
        return cls(spend_limit=Coins.from_data(data["spend_limit"]))

    def to_proto(self) -> SendAuthorization_pb:
        return SendAuthorization_pb(spend_limit=self.spend_limit.to_proto())

    @classmethod
    def from_proto(cls, proto: SendAuthorization_pb) -> SendAuthorization:
        return cls(spend_limit=Coins.from_proto(proto.spend_limit))

    @classmethod
    def from_amino(cls, amino: dict) -> SendAuthorization:
        value = amino["value"]
        return cls(spend_limit=Coins.from_amino(value["spend_limit"]))

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))


@attr.s
class GenericAuthorization(Authorization):
    """Generic type of :class:`Authorization`, specifying the type of message to allow.

    Args:
        msg: type of message allowed by authorization"""

    type_amino = "cosmos-sdk/GenericAuthorization"
    """"""
    type_url = "/cosmos.authz.v1beta1.GenericAuthorization"
    """"""
    prototype = GenericAuthorization_pb
    """"""

    msg: str = attr.ib()

    def to_amino(self) -> dict:
        return {"type": self.type_amino, "value": {"msg": self.msg}}

    def to_data(self) -> dict:
        return {"@type": self.type_url, "msg": self.msg}

    @classmethod
    def from_data(cls, data: dict) -> GenericAuthorization:
        return cls(msg=data["msg"])

    def to_proto(self) -> GenericAuthorization_pb:
        return GenericAuthorization_pb(msg=self.msg)

    @classmethod
    def from_proto(cls, proto: GenericAuthorization_pb) -> GenericAuthorization:
        return cls(msg=proto.msg)

    @classmethod
    def from_amino(cls, amino: dict) -> GenericAuthorization:
        value = amino["value"]
        return cls(msg=value["msg"])

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))


@attr.s
class AuthorizationGrant(JSONSerializable):
    """Contains information about an existing granted authorization between two users."""

    authorization: Authorization = attr.ib()
    """Grant authorization details."""

    expiration: datetime = attr.ib(converter=parser.parse)
    """Grant expiration."""

    def to_amino(self) -> dict:
        return {
            "authorization": self.authorization.to_amino(),
            "expiration": to_isoformat(self.expiration),
        }

    def to_data(self) -> dict:
        return {
            "authorization": self.authorization.to_data(),
            "expiration": to_isoformat(self.expiration),
        }

    @classmethod
    def from_data(cls, data: dict) -> AuthorizationGrant:
        return cls(
            authorization=Authorization.from_data(data["authorization"]),
            expiration=data["expiration"],
        )

    def to_proto(self) -> Grant_pb:
        return Grant_pb(
            authorization=self.authorization.pack_any(),
            expiration=self.expiration,
        )

    @classmethod
    def from_proto(cls, proto: Grant_pb) -> AuthorizationGrant:
        return cls(
            authorization=Authorization.unpack_any(proto.authorization),
            expiration=proto.expiration,
        )

    @classmethod
    def from_amino(cls, amino: dict) -> AuthorizationGrant:
        value = amino["value"]
        return cls(
            authorization=Authorization.from_amino(value["authorization"]),
            expiration=value["expiration"],
        )


@attr.s
class StakeAuthorizationValidators(JSONSerializable):
    address: List[AccAddress] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    def to_data(self) -> dict:
        return {"address": self.address}

    @classmethod
    def from_data(cls, data: dict) -> StakeAuthorizationValidators:
        return cls(address=data["address"])

    def to_proto(self):
        return StakeAuthorizationValidators_pb(address=self.address)

    @classmethod
    def from_proto(
        cls, proto: StakeAuthorizationValidators_pb
    ) -> StakeAuthorizationValidators:
        return cls(address=proto.address)


@attr.s
class StakeAuthorization(Authorization):
    authorization_type: AuthorizationType = attr.ib()
    max_tokens: Optional[Coin] = attr.ib(default=None)
    allow_list: Optional[StakeAuthorizationValidators] = attr.ib(default=None)
    deny_list: Optional[StakeAuthorizationValidators] = attr.ib(default=None)

    type_url = "/cosmos.staking.v1beta1.StakeAuthorization"
    """"""
    prototype = StakeAuthorization_pb
    """"""

    def to_amino(self):
        raise Exception("Amino not supported")

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "authorization_type": self.authorization_type,
            "max_tokens": self.max_tokens.to_data() if self.max_tokens else None,
            "allow_list": self.allow_list.to_data() if self.allow_list else None,
            "deny_list": self.deny_list.to_data() if self.deny_list else None,
        }

    @classmethod
    def from_data(cls, data: dict) -> StakeAuthorization:
        return StakeAuthorization(
            authorization_type=data["authorization_type"],
            max_tokens=(
                Coin.from_data(data["max_tokens"])
                if data.get("max_tokens") is not None
                else None
            ),
            allow_list=StakeAuthorizationValidators.from_data(data["allow_list"])
            if data.get("allow_list")
            else None,
            deny_list=StakeAuthorizationValidators.from_data(data["deny_list"])
            if data.get("deny_list")
            else None,
        )

    def to_proto(self) -> StakeAuthorization_pb:
        return StakeAuthorization_pb(
            authorization_type=self.authorization_type,
            max_tokens=self.max_tokens.to_proto() if self.max_tokens else None,
            allow_list=self.allow_list.to_proto() if self.allow_list else None,
            deny_list=self.deny_list.to_proto() if self.deny_list else None,
        )

    @classmethod
    def from_proto(cls, proto: StakeAuthorization_pb) -> StakeAuthorization:
        return StakeAuthorization(
            authorization_type=proto.authorization_type,
            max_tokens=Coin.from_proto(proto.max_tokens) if proto.max_tokens else None,
            allow_list=StakeAuthorizationValidators.from_proto(proto.allow_list)
            if proto.allow_list
            else None,
            deny_list=StakeAuthorizationValidators.from_proto(proto.deny_list)
            if proto.deny_list
            else None,
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))



from functools import reduce
from math import ceil
from typing import Any, Dict, Union

from terra_sdk.core import Coin

from .api._base import BaseAsyncAPI, sync_bind


def index_by_pub_key(m: Dict[str, Any], o: Any):
    m[o["pub_key"]["key"]] = o
    return m


class AsyncLCDUtils(BaseAsyncAPI):
    async def validators_with_voting_power(self) -> Dict[str, dict]:
        """Gets current validators and merges their voting power from the validator set query.

        Returns:
            Dict[str, dict]: validators with voting power
        """
        validator_set_response = await BaseAsyncAPI._try_await(
            self._c.tendermint.validator_set()
        )
        next_key = ""
        while True:
            from terra_sdk.client.lcd import PaginationOptions

            validators, pag = await BaseAsyncAPI._try_await(
                self._c.staking.validators(PaginationOptions(key=next_key))
            )
            validator_set: Dict[str, Any] = reduce(
                index_by_pub_key, validator_set_response["validators"], {}
            )
            if pag is None or pag["next_key"] is None:
                break
            next_key = pag["next_key"]
        res = {}
        for v in validators:
            delegate_info = validator_set.get(v.consensus_pubkey["key"])
            if delegate_info is None:
                continue
            res[v.operator_address] = {
                "validator_info": v,
                "voting_power": int(delegate_info["voting_power"]),
                "proposer_priority": int(delegate_info["proposer_priority"]),
            }
        return res


class LCDUtils(AsyncLCDUtils):
    @sync_bind(AsyncLCDUtils.validators_with_voting_power)
    async def validators_with_voting_power(self) -> Dict[str, dict]:
        pass



from .channel import Channel, Counterparty, Order, Packet, State
from .client import (
    ClientConsensusStates,
    ConsensusStateWithHeight,
    Height,
    IdentifiedClientState,
    Params,
)
from .commitment import MerklePrefix, MerkleRoot

__all__ = [
    "Height",
    "IdentifiedClientState",
    "ConsensusStateWithHeight",
    "ClientConsensusStates",
    "Params",
    "MerkleRoot",
    "MerklePrefix",
    "Counterparty",
    "Channel",
    "Order",
    "State",
    "Packet",
]



from __future__ import annotations

import json
from typing import List, Optional, Union

import attr
from terra_proto.cosmwasm.wasm.v1 import AbsoluteTxPosition as AbsoluteTxPosition_pb
from terra_proto.cosmwasm.wasm.v1 import AccessConfig as AccessConfig_pb
from terra_proto.cosmwasm.wasm.v1 import AccessConfigUpdate as AccessConfigUpdate_pb
from terra_proto.cosmwasm.wasm.v1 import AccessType
from terra_proto.cosmwasm.wasm.v1 import AccessTypeParam as AccessTypeParam_pb
from terra_proto.cosmwasm.wasm.v1 import ContractCodeHistoryEntry as HistoryEntry_pb
from terra_proto.cosmwasm.wasm.v1 import ContractCodeHistoryOperationType

from terra_sdk.core.bech32 import AccAddress
from terra_sdk.core.msg import Msg
from terra_sdk.util.json import JSONSerializable

__all__ = ["AccessType", "AccessConfig", "AccessConfigUpdate", "AccessTypeParam"]


def parse_msg(msg: Union[dict, str, bytes]) -> dict:
    if type(msg) is dict:
        return msg
    return json.loads(msg)


def convert_access_type_from_json(access_type: str) -> AccessType:
    if access_type == "Everybody":
        return AccessType.ACCESS_TYPE_EVERYBODY
    elif access_type == "Nobody":
        return AccessType.ACCESS_TYPE_NOBODY
    elif access_type == "OnlyAddress":
        return AccessType.ACCESS_TYPE_ONLY_ADDRESS
    elif access_type == "Unspecified":
        return AccessType.ACCESS_TYPE_UNSPECIFIED


def convert_access_type_to_json(access_type: AccessType) -> str:
    if access_type == AccessType.ACCESS_TYPE_EVERYBODY:
        return AccessType.ACCESS_TYPE_EVERYBODY.name
    elif access_type == AccessType.ACCESS_TYPE_NOBODY:
        return AccessType.ACCESS_TYPE_NOBODY.name
    elif access_type == AccessType.ACCESS_TYPE_ONLY_ADDRESS:
        return AccessType.ACCESS_TYPE_ONLY_ADDRESS.name
    elif access_type == AccessType.ACCESS_TYPE_UNSPECIFIED:
        return AccessType.ACCESS_TYPE_UNSPECIFIED.name


def convert_history_operation_type_to_json(
    operation_type: str,
) -> ContractCodeHistoryOperationType:
    if (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_GENESIS
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_GENESIS.name
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_INIT
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_INIT.name
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_MIGRATE
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_MIGRATE.name
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_UNSPECIFIED
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_UNSPECIFIED.name
        )


def convert_history_operation_type_from_json(
    operation_type: ContractCodeHistoryOperationType,
) -> str:
    if (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_GENESIS.name
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_GENESIS
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_INIT.name
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_INIT
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_MIGRATE.name
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_MIGRATE
        )
    elif (
        operation_type
        == ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_UNSPECIFIED.name
    ):
        return (
            ContractCodeHistoryOperationType.CONTRACT_CODE_HISTORY_OPERATION_TYPE_UNSPECIFIED
        )


@attr.s
class AccessConfig(JSONSerializable):

    permission: AccessType = attr.ib()
    """"""

    address: AccAddress = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "permission": convert_access_type_to_json(self.permission),
            "address": self.address,
        }

    def to_data(self) -> dict:
        return {
            "permission": convert_access_type_to_json(self.permission),
            "address": self.address,
        }

    def to_proto(self) -> AccessConfig_pb:
        return AccessConfig_pb(permission=self.permission, address=self.address)

    @classmethod
    def from_data(cls, data: dict) -> AccessConfig:
        return cls(
            permission=convert_access_type_from_json(data["permission"]),
            address=data["address"],
        )

    @classmethod
    def from_proto(cls, proto: AccessConfig_pb) -> AccessConfig:
        return cls(permission=proto.permission, address=proto.address)


@attr.s
class AccessConfigUpdate(JSONSerializable):

    code_id: int = attr.ib()
    """"""

    instantiate_permission: AccessConfig = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "code_id": self.code_id,
            "instantiate_permission": self.instantiate_permission.to_amino(),
        }

    def to_proto(self) -> AccessConfigUpdate_pb:
        return AccessConfigUpdate_pb(
            code_id=self.code_id,
            instantiate_permission=self.instantiate_permission.to_proto(),
        )

    @classmethod
    def from_data(cls, data: dict) -> AccessConfigUpdate:
        return cls(
            code_id=data["code_id"],
            instantiate_permission=AccessConfig.from_data(
                data["instantiate_permission"]
            )
            if "instantiate_permission" in data
            else None,
        )

    @classmethod
    def from_proto(cls, proto: AccessConfigUpdate_pb) -> AccessConfigUpdate:
        return cls(
            code_id=proto.code_id,
            instantiate_permission=AccessConfig.from_proto(proto.instantiate_permission)
            if proto.instantiate_permission
            else None,
        )


@attr.s
class AccessTypeParam(JSONSerializable):
    value: AccessType = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "value": convert_access_type_to_json(self.value),
        }

    def to_proto(self) -> AccessTypeParam_pb:
        return AccessTypeParam_pb(
            value=self.value,
        )

    @classmethod
    def from_data(cls, data: dict) -> AccessTypeParam:
        return cls(
            value=convert_access_type_from_json(data["value"]),
        )

    @classmethod
    def from_proto(cls, proto: AccessTypeParam_pb) -> AccessTypeParam:
        return cls(value=proto.value)


@attr.s
class AbsoluteTxPosition(JSONSerializable):
    block_height: int = attr.ib()
    tx_index: int = attr.ib()
    """"""

    def to_amino(self):
        return {"block_height": self.block_height, "tx_index": self.tx_index}

    def to_proto(self) -> AbsoluteTxPosition_pb:
        return AbsoluteTxPosition_pb(
            block_height=self.block_height, tx_index=self.tx_index
        )

    @classmethod
    def from_data(cls, data: dict) -> AbsoluteTxPosition:
        return cls(block_height=data["block_height"], tx_index=data["tx_index"])

    @classmethod
    def from_proto(cls, proto: AbsoluteTxPosition_pb) -> AbsoluteTxPosition:
        return cls(block_height=proto.block_height, tx_index=proto.tx_index)


@attr.s
class HistoryEntry(JSONSerializable):
    operation: ContractCodeHistoryOperationType = attr.ib()
    code_id: int = attr.ib()
    updated: Optional[AbsoluteTxPosition] = attr.ib()
    msg: Union[str, dict] = attr.ib()
    """"""

    def to_amino(self):
        return {
            "operation": convert_history_operation_type_to_json(self.operation),
            "code_id": self.code_id,
            "updated": self.updated.to_amino(),
            "msg": self.msg,
        }

    def to_data(self):
        return {
            "operation": convert_history_operation_type_to_json(self.operation),
            "code_id": self.code_id,
            "updated": self.updated.to_data(),
            "msg": self.msg,
        }

    def to_proto(self) -> HistoryEntry_pb:
        return HistoryEntry_pb(
            operation=self.operation,
            code_id=self.code_id,
            updated=self.updated,
            msg=bytes(json.dumps(self.msg), "utf-8"),
        )

    @classmethod
    def from_data(cls, data: dict) -> HistoryEntry:
        return cls(
            operation=convert_history_operation_type_from_json(data["operation"]),
            code_id=data["code_id"],
            updated=AbsoluteTxPosition.from_data(data["updated"])
            if data["updated"]
            else None,
            msg=data["msg"],
        )

    @classmethod
    def from_proto(cls, proto: HistoryEntry_pb) -> HistoryEntry:
        return cls(
            operation=proto.operation,
            code_id=proto.code_id,
            updated=AbsoluteTxPosition.from_proto(proto.updated),
            msg=parse_msg(proto.msg),
        )



from typing import List, Optional

from terra_sdk.core import AccAddress
from terra_sdk.core.authz import AuthorizationGrant

from ..params import APIParams
from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncAuthzAPI", "AuthzAPI"]


class AsyncAuthzAPI(BaseAsyncAPI):
    async def grants(
        self,
        granter: AccAddress,
        grantee: AccAddress,
        msg_type: Optional[str] = None,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        """Fetches current active message authorization grants.

        Args:
            granter (AccAddress): granter account address
            grantee (AccAddress): grantee account address
            msg_type (str, optional): message type.
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[AuthorizationGrant]: message authorization grants matching criteria
        """
        params = {
            "granter": granter,
            "grantee": grantee,
        }
        if msg_type is not None:
            params["msg_type_url"] = msg_type

        res = await self._c._get("/cosmos/authz/v1beta1/grants", params)
        return [AuthorizationGrant.from_data(x) for x in res["grants"]]

    async def granter(
        self,
        granter: AccAddress,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        """Fetches list of `GrantAuthorization` granted by granter.

        Args:
            granter (AccAddress): granter account address
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[AuthorizationGrant]: message authorization grants matching criteria
        """

        res = await self._c._get(
            f"/cosmos/authz/v1beta1/grants/granter/{granter}", params
        )
        return [AuthorizationGrant.from_data(x) for x in res["grants"]]

    async def grantee(
        self,
        grantee: AccAddress,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        """Fetches list of `GrantAuthorization` by grantee.

        Args:
            grantee (AccAddress): grantee account address
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[AuthorizationGrant]: message authorization grants matching criteria
        """

        res = await self._c._get(
            f"/cosmos/authz/v1beta1/grants/grantee/{grantee}", params
        )
        return [AuthorizationGrant.from_data(x) for x in res["grants"]]


class AuthzAPI(AsyncAuthzAPI):
    @sync_bind(AsyncAuthzAPI.grants)
    def grants(
        self,
        granter: AccAddress,
        grantee: AccAddress,
        msg_type: Optional[str] = None,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        pass

    @sync_bind(AsyncAuthzAPI.granter)
    def granter(
        self,
        granter: AccAddress,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        pass

    @sync_bind(AsyncAuthzAPI.grantee)
    def grantee(
        self,
        grantee: AccAddress,
        params: Optional[APIParams] = None,
    ) -> List[AuthorizationGrant]:
        pass

    grants.__doc__ = AsyncAuthzAPI.grants.__doc__
    granter.__doc__ = AsyncAuthzAPI.granter.__doc__
    grantee.__doc__ = AsyncAuthzAPI.grantee.__doc__



"Wasm module messages."

from __future__ import annotations

import base64
import json
from cProfile import label
from typing import Optional, Union

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmwasm.wasm.v1 import MsgClearAdmin as MsgClearAdmin_pb
from terra_proto.cosmwasm.wasm.v1 import MsgExecuteContract as MsgExecuteContract_pb
from terra_proto.cosmwasm.wasm.v1 import (
    MsgInstantiateContract as MsgInstantiateContract_pb,
)
from terra_proto.cosmwasm.wasm.v1 import MsgMigrateContract as MsgMigrateContract_pb
from terra_proto.cosmwasm.wasm.v1 import MsgStoreCode as MsgStoreCode_pb
from terra_proto.cosmwasm.wasm.v1 import MsgUpdateAdmin as MsgUpdateAdmin_pb

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg
from terra_sdk.core.wasm.data import AccessConfig, AccessTypeParam
from terra_sdk.util.remove_none import remove_none

__all__ = [
    "MsgStoreCode",
    "MsgInstantiateContract",
    "MsgExecuteContract",
    "MsgMigrateContract",
    "MsgUpdateAdmin",
    "MsgClearAdmin",
]


def parse_msg(msg: Union[dict, str, bytes]) -> dict:
    if type(msg) is dict:
        return msg
    return json.loads(msg)


@attr.s
class MsgStoreCode(Msg):
    """Upload a new smart contract WASM binary to the blockchain.

    Args:
        sender: address of sender
        wasm_byte_code: base64-encoded string containing bytecode
        instantiate_permission: access control to apply on contract creation, optional
    """

    type_amino = "wasm/MsgStoreCode"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgStoreCode"
    """"""
    prototype = MsgStoreCode_pb
    """"""

    sender: AccAddress = attr.ib()
    wasm_byte_code: str = attr.ib()
    instantiate_permission: AccessConfig = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "wasm_byte_code": self.wasm_byte_code,
                "instantiate_permission": self.instantiate_permission.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgStoreCode:
        return cls(
            sender=data["sender"],
            wasm_byte_code=data["wasm_byte_code"],
            instantiate_permission=AccessConfig.from_data(
                data["instantiate_permission"]
            ),
        )

    def to_proto(self) -> MsgStoreCode_pb:
        return MsgStoreCode_pb(
            sender=self.sender,
            wasm_byte_code=base64.b64decode(self.wasm_byte_code),
            instantiate_permission=self.instantiate_permission.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgStoreCode_pb) -> MsgStoreCode:
        return cls(
            sender=proto.sender,
            wasm_byte_code=base64.b64encode(proto.wasm_byte_code).decode(),
            instantiate_permission=AccessConfig.from_proto(
                proto.instantiate_permission
            ),
        )


@attr.s
class MsgInstantiateContract(Msg):
    """Creates a new instance of a smart contract from existing code on the blockchain.

    Args:
        sender: address of sender
        admin: address of contract admin
        code_id (int): code ID to use for instantiation
        label (str): label for the contract.
        msg (dict|str): InitMsg to initialize contract
        funds (Coins): initial amount of coins to be sent to contract
    """

    type_amino = "wasm/MsgInstantiateContract"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgInstantiateContract"
    """"""
    prototype = MsgInstantiateContract_pb
    """"""

    sender: AccAddress = attr.ib()
    admin: Optional[AccAddress] = attr.ib()
    code_id: int = attr.ib(converter=int)
    label: str = attr.ib(converter=str)
    msg: Union[dict, str] = attr.ib()
    funds: Coins = attr.ib(converter=Coins, factory=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "admin": self.admin,
                "code_id": str(self.code_id),
                "label": self.label,
                "msg": remove_none(self.msg),
                "funds": self.funds.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgInstantiateContract:
        return cls(
            sender=data.get("sender"),
            admin=data.get("admin"),
            code_id=data["code_id"],
            label=data["label"],
            msg=parse_msg(data["msg"]),
            funds=Coins.from_data(data["funds"]),
        )

    def to_proto(self) -> MsgInstantiateContract_pb:
        return MsgInstantiateContract_pb(
            sender=self.sender,
            admin=self.admin,
            code_id=self.code_id,
            label=self.label,
            msg=bytes(json.dumps(self.msg), "utf-8"),
            funds=self.funds.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgInstantiateContract_pb) -> MsgInstantiateContract:
        return cls(
            sender=proto.sender,
            admin=proto.admin,
            code_id=proto.code_id,
            label=proto.label,
            msg=parse_msg(proto.msg),
            funds=Coins.from_proto(proto.funds),
        )


@attr.s
class MsgExecuteContract(Msg):
    """Execute a state-mutating function on a smart contract.

    Args:
        sender: address of sender
        contract: address of contract to execute function on
        msg (dict|str): ExecuteMsg to pass
        coins: coins to be sent, if needed by contract to execute.
            Defaults to empty ``Coins()``
    """

    type_amino = "wasm/MsgExecuteContract"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgExecuteContract"
    """"""
    prototype = MsgExecuteContract_pb
    """"""

    sender: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()
    msg: Union[dict, str] = attr.ib()
    coins: Coins = attr.ib(converter=Coins, factory=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "contract": self.contract,
                "msg": remove_none(self.msg),
                "coins": self.coins.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgExecuteContract:
        return cls(
            sender=data["sender"],
            contract=data["contract"],
            msg=parse_msg(data["msg"]),
            coins=Coins.from_data(data["coins"]),
        )

    def to_proto(self) -> MsgExecuteContract_pb:
        return MsgExecuteContract_pb(
            sender=self.sender,
            contract=self.contract,
            msg=bytes(json.dumps(self.msg), "utf-8"),
            funds=self.coins.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgExecuteContract_pb) -> MsgExecuteContract:
        return cls(
            sender=proto.sender,
            contract=proto.contract,
            msg=parse_msg(proto.msg),
            coins=(proto.funds),
        )


@attr.s
class MsgMigrateContract(Msg):
    """Migrate the contract to a different code ID.

    Args:
        sender: address of contract admin
        contract: address of contract to migrate
        code_id (int): new code ID to migrate to
        msg (dict|str): MigrateMsg to execute
    """

    type_amino = "wasm/MsgMigrateContract"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgMigrateContract"
    """"""
    prototype = MsgMigrateContract_pb
    """"""

    sender: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()
    code_id: int = attr.ib(converter=int)
    msg: Union[dict, str] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "contract": self.contract,
                "code_id": str(self.code_id),
                "msg": remove_none(self.msg),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgMigrateContract:
        return cls(
            sender=data["sender"],
            contract=data["contract"],
            code_id=data["code_id"],
            msg=parse_msg(data["msg"]),
        )

    def to_proto(self) -> MsgMigrateContract_pb:
        return MsgMigrateContract_pb(
            sender=self.sender,
            contract=self.contract,
            code_id=self.code_id,
            msg=bytes(json.dumps(self.msg), "utf-8"),
        )

    @classmethod
    def from_proto(cls, proto: MsgMigrateContract_pb) -> MsgMigrateContract:
        return cls(
            sender=proto.sender,
            contract=proto.contract,
            code_id=proto.code_id,
            msg=parse_msg(proto.msg),
        )


@attr.s
class MsgUpdateAdmin(Msg):
    """Update a smart contract's admin.

    Args:
        sender: address of current admin (sender)
        new_admin: address of new admin
        contract: address of contract to change
    """

    type_amino = "wasm/MsgUpdateAdmin"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgUpdateAdmin"
    """"""
    prototype = MsgUpdateAdmin_pb
    """"""

    sender: AccAddress = attr.ib()
    new_admin: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "new_admin": self.new_admin,
                "contract": self.contract,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgUpdateAdmin:
        return cls(
            sender=data["sender"],
            new_admin=data["new_admin"],
            contract=data["contract"],
        )

    def to_proto(self) -> MsgUpdateAdmin_pb:
        return MsgUpdateAdmin_pb(
            sender=self.sender, new_admin=self.new_admin, contract=self.contract
        )

    @classmethod
    def from_proto(cls, proto: MsgUpdateAdmin_pb) -> MsgUpdateAdmin:
        return cls(
            sender=proto.sender,
            new_admin=proto.new_admin,
            contract=proto.contract,
        )


@attr.s
class MsgClearAdmin(Msg):
    """Clears the contract's admin field.

    Args:
        admin: address of current admin (sender)
        contract: address of contract to change
    """

    type_amino = "wasm/MsgClearAdmin"
    """"""
    type_url = "/cosmwasm.wasm.v1.MsgClearAdmin"
    """"""
    prototype = MsgClearAdmin_pb
    """"""

    sender: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"sender": self.sender, "contract": self.contract},
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgClearAdmin:
        return cls(
            sender=data["sender"],
            contract=data["contract"],
        )

    def to_proto(self) -> MsgClearAdmin_pb:
        return MsgClearAdmin_pb(sender=self.sender, contract=self.contract)

    @classmethod
    def from_proto(cls, proto: MsgClearAdmin_pb) -> MsgClearAdmin:
        return cls(
            sender=proto.sender,
            contract=proto.contract,
        )



"""Data objects about Signing Mode Info."""

from __future__ import annotations

from typing import List, Optional

import attr
from terra_proto.cosmos.tx.v1beta1 import ModeInfo as ModeInfo_pb
from terra_proto.cosmos.tx.v1beta1 import ModeInfoMulti as ModeInfoMulti_pb
from terra_proto.cosmos.tx.v1beta1 import ModeInfoSingle as ModeInfoSingle_pb

from terra_sdk.util.json import JSONSerializable

from .compact_bit_array import CompactBitArray

__all__ = [
    "ModeInfo",
    "ModeInfoSingle",
    "ModeInfoMulti",
]

from terra_proto.cosmos.tx.signing.v1beta1 import SignMode


@attr.s
class ModeInfo(JSONSerializable):

    single: Optional[ModeInfoSingle] = attr.ib(default=None)
    multi: Optional[ModeInfoMulti] = attr.ib(default=None)

    def to_data(self) -> dict:
        if self.single:
            return {"single": self.single.to_data()}
        if self.multi:
            return {"multi": self.multi.to_data()}
        raise ValueError("ModeInfo should have one of single or multi")

    @classmethod
    def from_data(cls, data: dict) -> ModeInfo:
        if data.get("single"):
            return cls(single=ModeInfoSingle.from_data(data.get("single")))
        if data.get("multi"):
            return cls(multi=ModeInfoMulti.from_data(data.get("multi")))
        raise ValueError("ModeInfo should have one of single or multi")

    def to_proto(self) -> ModeInfo_pb:
        if self.single:
            return ModeInfo_pb(single=self.single.to_proto())
        else:
            return ModeInfo_pb(multi=(self.multi.to_proto() if self.multi else None))

    @classmethod
    def from_proto(cls, proto: ModeInfo_pb) -> ModeInfo:
        if proto.single is not None:
            return ModeInfo(single=ModeInfoSingle.from_proto(proto.single))
        else:
            return ModeInfo(multi=ModeInfoMulti.from_proto(proto.multi))


@attr.s
class ModeInfoSingle(JSONSerializable):
    mode: SignMode = attr.ib()

    def to_data(self) -> dict:
        return {"mode": self.mode.name}

    @classmethod
    def from_data(cls, data: dict) -> ModeInfoSingle:
        return cls(SignMode[data["mode"]])

    def to_proto(self) -> ModeInfoSingle_pb:
        return ModeInfoSingle_pb(mode=self.mode)

    @classmethod
    def from_proto(cls, proto: ModeInfoSingle_pb) -> ModeInfoSingle:
        mode = SignMode(proto.mode)
        return cls(mode=mode)


@attr.s
class ModeInfoMulti(JSONSerializable):
    bitarray: CompactBitArray = attr.ib()
    mode_infos: List[ModeInfo] = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> ModeInfoMulti:
        return cls(
            CompactBitArray.from_data(data["bitarray"]),
            [ModeInfo.from_data(d) for d in data["mode_infos"]],
        )

    def to_data(self) -> dict:
        return {
            "bitarray": self.bitarray.to_data(),
            "mode_infos": [mi.to_data() for mi in self.mode_infos],
        }

    def to_proto(self) -> ModeInfoMulti_pb:
        return ModeInfoMulti_pb(
            bitarray=self.bitarray.to_proto(),
            mode_infos=[mi.to_proto() for mi in self.mode_infos],
        )

    @classmethod
    def from_proto(cls, proto: ModeInfoMulti_pb) -> ModeInfoMulti:
        return cls(
            CompactBitArray.from_proto(proto.bitarray),
            ModeInfo_pb.from_proto(proto.mode_infos),
        )



"""Upgrade module data objects."""

from __future__ import annotations

__all__ = ["Plan"]

from datetime import datetime
from typing import Any, Optional

import attr
from attr import converters
from betterproto.lib.google.protobuf import Any as Any_pb
from dateutil import parser
from terra_proto.cosmos.upgrade.v1beta1 import Plan as Plan_pb

from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable


@attr.s
class Plan(JSONSerializable):
    name: str = attr.ib()
    height: str = attr.ib()
    info: str = attr.ib()
    time: Optional[datetime] = attr.ib(
        default=None, converter=converters.optional(parser.parse)
    )
    upgrade_client_state: Optional[Any] = attr.ib(default=None)

    def to_amino(self) -> dict:
        return {
            "name": self.name,
            "height": self.height,
            "info": self.info,
            "time": to_isoformat(self.time) if self.time else None,
            "upgrade_client_state": self.upgrade_client_state,
        }

    @classmethod
    def from_data(cls, data: dict) -> Plan:
        return cls(
            name=data["name"],
            time=data["time"] if data.get("time") else None,
            height=data["height"],
            info=data["info"],
            upgrade_client_state=data["upgrade_client_state"]
            if data.get("upgrade_client_state")
            else None,
        )

    def to_proto(self) -> Plan_pb:
        ucs = self.upgrade_client_state
        if ucs is not None:
            ucs = Any_pb(type_url=ucs["type_url"], value=bytes(ucs.to_proto()))
        return Plan_pb(
            name=self.name,
            time=self.time,
            height=self.height,
            info=self.info,
            upgraded_client_state=ucs if ucs else None,
        )

    @classmethod
    def from_proto(cls, proto: Plan_pb) -> Plan:
        return cls(
            name=proto.name,
            time=proto.time if proto.time else None,
            height=proto.height,
            info=proto.info,
            upgrade_client_state=Any_pb().parse(proto.upgrade_client_state)
            if proto.upgraded_client_state
            else None,
        )



"""Distribution module message types."""

from __future__ import annotations

import attr
from terra_proto.cosmos.distribution.v1beta1 import (
    MsgFundCommunityPool as MsgFundCommunityPool_pb,
)
from terra_proto.cosmos.distribution.v1beta1 import (
    MsgSetWithdrawAddress as MsgSetWithdrawAddress_pb,
)
from terra_proto.cosmos.distribution.v1beta1 import (
    MsgWithdrawDelegatorReward as MsgWithdrawDelegatorReward_pb,
)
from terra_proto.cosmos.distribution.v1beta1 import (
    MsgWithdrawValidatorCommission as MsgWithdrawValidatorCommission_pb,
)

from terra_sdk.core import AccAddress, Coins, ValAddress
from terra_sdk.core.msg import Msg

__all__ = [
    "MsgSetWithdrawAddress",
    "MsgWithdrawDelegatorReward",
    "MsgWithdrawValidatorCommission",
    "MsgFundCommunityPool",
]


@attr.s
class MsgSetWithdrawAddress(Msg):
    """Modify Withdraw Address of a delegator.

    Args:
        delegator_address: delegator
        withdraw_address: new withdraw address
    """

    type_amino = "cosmos-sdk/MsgModifyWithdrawAddress"
    """"""
    type_url = "/cosmos.distribution.v1beta1.MsgSetWithdrawAddress"
    """"""
    action = "set_withdraw_address"
    """"""
    prototype = MsgSetWithdrawAddress_pb
    """"""

    delegator_address: AccAddress = attr.ib()
    withdraw_address: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "delegator_address": self.delegator_address,
                "withdraw_address": self.withdraw_address,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "delegator_address": self.delegator_address,
            "withdraw_address": self.withdraw_address,
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgSetWithdrawAddress:
        return cls(
            delegator_address=data["delegator_address"],
            withdraw_address=data["withdraw_address"],
        )

    def to_proto(self) -> MsgSetWithdrawAddress_pb:
        return MsgSetWithdrawAddress_pb(
            delegator_address=self.delegator_address,
            withdraw_address=self.withdraw_address,
        )

    @classmethod
    def from_proto(cls, data: MsgSetWithdrawAddress_pb) -> MsgSetWithdrawAddress:
        return cls(
            delegator_address=data["delegator_address"],
            withdraw_address=data["withdraw_address"],
        )


@attr.s
class MsgWithdrawDelegatorReward(Msg):
    """Withdraw rewards for a delegation specified by a (delegator, validator) pair.

    Args:
        delegator_address: delegator
        validator_address: validator
    """

    type_amino = "cosmos-sdk/MsgWithdrawDelegationReward"
    """"""
    type_url = "/cosmos.distribution.v1beta1.MsgWithdrawDelegatorReward"
    """"""
    action = "withdraw_delegation_reward"
    """"""
    prototype = MsgWithdrawDelegatorReward_pb
    """"""

    delegator_address: AccAddress = attr.ib()
    validator_address: ValAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "delegator_address": self.delegator_address,
            "validator_address": self.validator_address,
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgWithdrawDelegatorReward:
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
        )

    def to_proto(self) -> MsgWithdrawDelegatorReward_pb:
        return MsgWithdrawDelegatorReward_pb(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
        )

    @classmethod
    def from_proto(
        cls, proto: MsgWithdrawDelegatorReward_pb
    ) -> MsgWithdrawDelegatorReward:
        return cls(
            delegator_address=proto.delegator_address,
            validator_address=proto.validator_address,
        )


@attr.s
class MsgWithdrawValidatorCommission(Msg):
    """Withdraw rewards accrued for a validator gained through commissions.

    Args:
        validator_address: validator operator address
    """

    type_amino = "cosmos-sdk/MsgWithdrawValidatorCommission"
    """"""
    type_url = "/cosmos.distribution.v1beta1.MsgWithdrawValidatorCommission"
    """"""
    action = "withdraw_validator_commission"
    """"""
    prototype = MsgWithdrawValidatorCommission_pb
    """"""

    validator_address: ValAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"validator_address": self.validator_address},
        }

    def to_data(self) -> dict:
        return {"@type": self.type_url, "validator_address": self.validator_address}

    @classmethod
    def from_data(cls, data: dict) -> MsgWithdrawValidatorCommission:
        return cls(validator_address=data["validator_address"])

    def to_proto(self) -> MsgWithdrawValidatorCommission_pb:
        return MsgWithdrawValidatorCommission_pb(
            validator_address=self.validator_address
        )

    @classmethod
    def from_proto(
        cls, data: MsgWithdrawValidatorCommission_pb
    ) -> MsgWithdrawValidatorCommission:
        return cls(validator_address=data["validator_address"])


@attr.s
class MsgFundCommunityPool(Msg):
    """Deposit assets to the Community Pool.

    Args:
        depositor (AccAddress): sender
        amount (Coins): amount to fund community pool with
    """

    type_amino = "cosmos-sdk/MsgFundCommunityPool"
    """"""
    type_url = "/cosmos.distribution.v1beta1.MsgFundCommunityPool"
    """"""
    prototype = MsgFundCommunityPool_pb
    """"""

    depositor: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"depositor": self.depositor, "amount": self.amount.to_amino()},
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "depositor": self.depositor,
            "amount": self.amount.to_data(),
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgFundCommunityPool:
        return cls(depositor=data["depositor"], amount=Coins.from_data(data["amount"]))

    def to_proto(self) -> MsgFundCommunityPool_pb:
        return MsgFundCommunityPool_pb(
            depositor=self.depositor, amount=self.amount.to_proto()
        )

    @classmethod
    def from_proto(cls, data: MsgFundCommunityPool_pb) -> MsgFundCommunityPool:
        return cls(depositor=data["depositor"], amount=Coins.from_proto(data["amount"]))



from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncIbcTransferAPI", "IbcTransferAPI"]


class AsyncIbcTransferAPI(BaseAsyncAPI):
    async def parameters(self) -> dict:
        """Fetches the IbcTransfer module's parameters.

        Returns:
            dict: IbcTransfer module parameters
        """
        res = await self._c._get("/ibc/apps/transfer/v1/params")
        params = res["params"]
        return {
            "send_enabled": bool(params["send_enabled"]),
            "receive_enabled": bool(params["receive_enabled"]),
        }


class IbcTransferAPI(AsyncIbcTransferAPI):
    @sync_bind(AsyncIbcTransferAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncIbcTransferAPI.parameters.__doc__



"""Special Bech32 String Types"""

from __future__ import annotations

from typing import NewType

from bech32 import bech32_decode, bech32_encode, convertbits

__all__ = [
    "AccAddress",
    "ValAddress",
    "AccPubKey",
    "ValPubKey",
    "is_acc_address",
    "is_acc_pubkey",
    "is_val_address",
    "is_val_pubkey",
    "is_valcons_pubkey",
    "to_acc_address",
    "to_acc_pubkey",
    "to_val_address",
    "to_val_pubkey",
    "get_bech",
]


def get_bech(prefix: str, payload: str) -> str:
    data = convertbits(bytes.fromhex(payload), 8, 5)
    if data is None:
        raise ValueError(f"could not parse data: prefix {prefix}, payload {payload}")
    return bech32_encode(prefix, data)  # base64 -> base32


def check_prefix_and_length(prefix: str, data: str, length: int):
    vals = bech32_decode(data)
    return vals[0] == prefix and len(data) == length


AccAddress = NewType("AccAddress", str)
AccAddress.__doc__ = """Terra Bech32 Account Address -- type alias of str."""

ValAddress = NewType("ValAddress", str)
ValAddress.__doc__ = """Terra Bech32 Validator Operator Address -- type alias of str."""

AccPubKey = NewType("AccPubKey", str)
AccPubKey.__doc__ = """Terra Bech32 Account Address -- type alias of str."""

ValPubKey = NewType("ValPubKey", str)
ValPubKey.__doc__ = """Terra Bech32 Validator PubKey -- type alias of str."""

# ValConsPubKey = NewType("ValConsPubKey", str)
# ValConsPubKey.__doc__ = (
#  """Terra Bech32 Validator Conensus PubKey -- type alias of str."""
# )


def is_acc_address(data: str) -> bool:
    """Checks whether the given string is a properly formatted Terra account address.

    Args:
        data (str): string to check

    Returns:
        bool: whether the string is a proper account address
    """
    return check_prefix_and_length("terra", data, 44)


def to_acc_address(data: ValAddress) -> AccAddress:
    """Converts a validator operator address into an account address.

    Args:
        data (ValAddress): validator operator address

    Raises:
        ValueError: if provided string is not Bech32

    Returns:
        AccAddress: account address
    """
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return AccAddress(bech32_encode("terra", vals[1]))


def is_val_address(data: str) -> bool:
    """Checks whether the given string is a properly formatted Terra validator operator
    address.

    Args:
        data (str): string to check

    Returns:
        bool: whether the string is a proper validator address
    """
    return check_prefix_and_length("terravaloper", data, 51)


def to_val_address(data: AccAddress) -> ValAddress:
    """Converts an account address into a validator operator address.

    Args:
        data (AccAddress): account address

    Raises:
        ValueError: if provided string is not Bech32

    Returns:
        ValAddress: validator operator address
    """
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return ValAddress(bech32_encode("terravaloper", vals[1]))


def is_acc_pubkey(data: str) -> bool:
    """Checks whether the provided string is a properly formatted Terra account pubkey.

    Args:
        data (str): string to check

    Returns:
        bool: whether string is account pubkey
    """
    return check_prefix_and_length("terrapub", data, 76)


def to_acc_pubkey(data: ValPubKey) -> AccPubKey:
    """Converts a validator pubkey into an account pubkey.

    Args:
        data (ValPubKey): validator pubkey

    Raises:
        ValueError: if provided string is not Bech32

    Returns:
        AccPubKey: account pubkey
    """
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return AccPubKey(bech32_encode("terrapub", vals[1]))


def is_val_pubkey(data: str) -> bool:
    """Checks whether provided string is a properly formatted Terra validator pubkey.

    Args:
        data (str): string to check

    Returns:
        bool: whether string is validator pubkey
    """
    return check_prefix_and_length("terravaloperpub", data, 83)


def to_val_pubkey(data: AccPubKey) -> ValPubKey:
    """Converts an account pubkey into a validator pubkey.

    Args:
        data (AccPubKey): account pubkey

    Raises:
        ValueError: if provided string is not Bech32

    Returns:
        ValPubKey: validator pubkey
    """
    vals = bech32_decode(data)
    if vals[1] is None:
        raise ValueError(f"invalid bech32: {data}")
    return ValPubKey(bech32_encode("terravaloperpub", vals[1]))


def is_valcons_pubkey(data: str) -> bool:  # -> ValConsPubKey:
    """Checks whether provided string is a properly formatted Terra validator consensus
    pubkey.

    Args:
        data (str): string to check

    Returns:
        bool: whether string is validator consensus pubkey
    """
    return check_prefix_and_length("terravalconspub", data, 83)



from __future__ import annotations

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.ibc.core.client.v1 import (
    ClientUpdateProposal as ClientUpdateProposal_pb,
)

__all__ = ["ClientUpdateProposal"]

from terra_sdk.util.json import JSONSerializable


@attr.s
class ClientUpdateProposal(JSONSerializable):
    """Proposal that allows updating IBC clients. If it passes, the substitute
    client's latest consensus state is copied over to the subject client.

    """

    type_amino = "ibc/ClientUpdateProposal"
    """"""
    type_url = "/ibc.core.client.v1.ClientUpdateProposal"
    """"""
    prototype = ClientUpdateProposal_pb
    """"""

    title: str = attr.ib()
    description: str = attr.ib()
    subject_client_id: str = attr.ib()
    substitute_client_id: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "subject_client_id": self.subject_client_id,
                "substitute_client_id": self.substitute_client_id,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> ClientUpdateProposal:
        return cls(
            title=data["title"],
            description=data["description"],
            subject_client_id=data["subject_client_id"],
            substitute_client_id=data["substitute_client_id"],
        )

    def to_proto(self) -> ClientUpdateProposal_pb:
        return ClientUpdateProposal_pb(
            title=self.title,
            description=self.description,
            subject_client_id=self.subject_client_id,
            substitute_client_id=self.substitute_client_id,
        )

    @classmethod
    def from_proto(cls, proto: ClientUpdateProposal_pb) -> ClientUpdateProposal:
        return cls(
            title=proto.title,
            description=proto.description,
            subject_client_id=proto.subject_client_id,
            substitute_client_id=proto.substitute_client_id,
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))



"""Data objects pertaining to accounts."""

from __future__ import annotations

from typing import Optional

import attr
from terra_proto.cosmos.auth.v1beta1 import BaseAccount as BaseAccount_pb

from ....core import AccAddress
from ....util.json import JSONSerializable
from ...public_key import PublicKey

__all__ = ["BaseAccount"]


@attr.s
class BaseAccount(JSONSerializable):
    """Stores information about an account."""

    type_amino = "cosmos-sdk/BaseAccount"
    type_url = "/cosmos.auth.v1beta1.BaseAccount"

    address: AccAddress = attr.ib()
    """"""

    public_key: Optional[PublicKey] = attr.ib()
    """"""

    account_number: int = attr.ib(converter=int)
    """"""

    sequence: int = attr.ib(converter=int)
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "address": self.address,
                "public_key": self.public_key.to_amino() if self.public_key else None,
                "account_number": str(self.account_number),
                "sequence": str(self.sequence),
            },
        }

    @classmethod
    def from_amino(cls, amino: dict) -> BaseAccount:
        amino = amino["value"] if "value" in amino else amino
        return cls(
            address=amino["address"],
            public_key=PublicKey.from_amino(amino["public_key"])
            if amino["public_key"]
            else None,
            account_number=amino["account_number"],
            sequence=amino["sequence"],
        )

    def get_account_number(self) -> int:
        return self.account_number

    def get_sequence(self) -> int:
        return self.sequence

    def get_public_key(self) -> PublicKey:
        return self.public_key

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "address": self.address,
            "pub_key": self.public_key and self.public_key.to_data(),
            "account_number": str(self.account_number),
            "sequence": str(self.sequence),
        }

    @classmethod
    def from_data(cls, data: dict) -> BaseAccount:
        return cls(
            address=data["address"],
            public_key=data.get("pub_key") and PublicKey.from_data(data["pub_key"]),
            account_number=data.get("account_number") or 0,
            sequence=data.get("sequence") or 0,
        )

    @classmethod
    def from_proto(cls, proto: BaseAccount_pb) -> BaseAccount:
        return cls(
            address=proto.address,
            public_key=PublicKey.from_proto(proto.pub_key),
            account_number=proto.account_number,
            sequence=proto.sequence,
        )

    def to_proto(self) -> BaseAccount_pb:
        return BaseAccount_pb(
            address=self.address,
            pub_key=self.public_key.to_proto() if self.public_key else None,
            account_number=self.account_number,
            sequence=self.sequence,
        )



from __future__ import annotations

from bip32utils import BIP32_HARDEN, BIP32Key
from mnemonic import Mnemonic

from .raw import RawKey

__all__ = ["MnemonicKey", "LUNA_COIN_TYPE"]

LUNA_COIN_TYPE = 330


class MnemonicKey(RawKey):
    """A MnemonicKey derives a private key using a BIP39 mnemonic seed phrase, and provides key-derivation options based on the BIP44 HD path standard.

    .. note:: You can change ``coin_type`` to 118 to derive the key for a legacy Terra
        wallet (shares ``coin_type`` with ATOM).

    Args:
        mnemonic (str, optional): space-separated mnemonic seed phrase. If not provided,
            a 24-word mnemonic will be generated.
        account (int, optional): HD path parameter - account number.
        index (int, optional): HD path parameter - account index.
        coin_type (int, optional): HD path parameter - coin type.
    """

    mnemonic: str
    """Mnemonic key phrase associated with the account (space-separated)."""

    account: int
    """HD path parameter: account number."""

    index: int
    """HD path parameter: account index."""

    coin_type: int
    """HD path parameter: coin type"""

    @property
    def hd_path(self) -> str:
        """Returns the BIP32 HD path for key-derivation:

        ``m/44'/COIN_TYPE'/ACCOUNT'/0/INDEX'``

        Returns:
            str: full BIP32 HD path
        """
        return f"m/44'/{self.coin_type}'/{self.account}'/0/{self.index}"

    def __init__(
        self,
        mnemonic: str = None,
        account: int = 0,
        index: int = 0,
        coin_type: int = LUNA_COIN_TYPE,
    ):
        if mnemonic is None:
            mnemonic = Mnemonic("english").generate(256)
        seed = Mnemonic("english").to_seed(mnemonic)
        root = BIP32Key.fromEntropy(seed)
        # derive from hdpath
        child = (
            root.ChildKey(44 + BIP32_HARDEN)
            .ChildKey(coin_type + BIP32_HARDEN)
            .ChildKey(account + BIP32_HARDEN)
            .ChildKey(0)
            .ChildKey(index)
        )

        super().__init__(child.PrivateKey())
        self.mnemonic = mnemonic
        self.coin_type = coin_type
        self.account = account
        self.index = index



from typing import Dict

from terra_sdk.key.mnemonic import MnemonicKey

from .lcd import AsyncLCDClient, AsyncWallet, LCDClient, Wallet

__all__ = ["LOCALTERRA_MNEMONICS", "LocalTerra", "AsyncLocalTerra"]

LOCALTERRA_MNEMONICS = {
    "validator": "satisfy adjust timber high purchase tuition stool faith fine install that you unaware feed domain license impose boss human eager hat rent enjoy dawn",
    "test1": "notice oak worry limit wrap speak medal online prefer cluster roof addict wrist behave treat actual wasp year salad speed social layer crew genius",
    "test2": "quality vacuum heart guard buzz spike sight swarm shove special gym robust assume sudden deposit grid alcohol choice devote leader tilt noodle tide penalty",
    "test3": "symbol force gallery make bulk round subway violin worry mixture penalty kingdom boring survey tool fringe patrol sausage hard admit remember broken alien absorb",
    "test4": "bounce success option birth apple portion aunt rural episode solution hockey pencil lend session cause hedgehog slender journey system canvas decorate razor catch empty",
    "test5": "second render cat sing soup reward cluster island bench diet lumber grocery repeat balcony perfect diesel stumble piano distance caught occur example ozone loyal",
    "test6": "spatial forest elevator battle also spoon fun skirt flight initial nasty transfer glory palm drama gossip remove fan joke shove label dune debate quick",
    "test7": "noble width taxi input there patrol clown public spell aunt wish punch moment will misery eight excess arena pen turtle minimum grain vague inmate",
    "test8": "cream sport mango believe inhale text fish rely elegant below earth april wall rug ritual blossom cherry detail length blind digital proof identify ride",
    "test9": "index light average senior silent limit usual local involve delay update rack cause inmate wall render magnet common feature laundry exact casual resource hundred",
    "test10": "prefer forget visit mistake mixture feel eyebrow autumn shop pair address airport diesel street pass vague innocent poem method awful require hurry unhappy shoulder",
}

LOCALTERRA_DEFAULTS = {
    "url": "http://localhost:1317",
    "chain_id": "localterra",
    "gas_prices": {"uluna": "0.15"},
    "gas_adjustment": 1.75,
}


class AsyncLocalTerra(AsyncLCDClient):
    """An :class:`AsyncLCDClient` that comes preconfigured with the default settings for
    connecting to a LocalTerra node.
    """

    wallets: Dict[str, AsyncWallet]
    """Ready-to use :class:`Wallet` objects with LocalTerra default accounts."""

    def __init__(self, *args, **kwargs):
        options = {**LOCALTERRA_DEFAULTS, **kwargs}
        super().__init__(*args, **options)
        self.wallets = {
            wallet_name: self.wallet(
                MnemonicKey(mnemonic=LOCALTERRA_MNEMONICS[wallet_name])
            )
            for wallet_name in LOCALTERRA_MNEMONICS
        }


class LocalTerra(LCDClient):
    """A :class:`LCDClient` that comes preconfigured with the default settings for
    connecting to a LocalTerra node.
    """

    wallets: Dict[str, Wallet]
    """Ready-to use :class:`Wallet` objects with LocalTerra default accounts.

    >>> terra = LocalTerra()
    >>> terra.wallets['test1'].key.acc_address
    'terra1x46rqay4d3cssq8gxxvqz8xt6nwlz4td20k38v'
    """

    def __init__(self, *args, **kwargs):
        options = {**LOCALTERRA_DEFAULTS, **kwargs}
        super().__init__(*args, **options)
        self.wallets = {
            wallet_name: self.wallet(
                MnemonicKey(mnemonic=LOCALTERRA_MNEMONICS[wallet_name])
            )
            for wallet_name in LOCALTERRA_MNEMONICS
        }



"""Data objects about SignDoc."""

from __future__ import annotations

import json

import attr
from terra_proto.cosmos.tx.v1beta1 import SignDoc as SignDoc_pb

from terra_sdk.core.tx import AuthInfo, TxBody
from terra_sdk.util.json import JSONSerializable

__all__ = ["SignDoc"]

from terra_sdk.util.remove_none import remove_none


@attr.s
class SignDoc(JSONSerializable):
    chain_id: str = attr.ib()
    account_number: int = attr.ib(converter=int)
    sequence: int = attr.ib(converter=int)
    auth_info: AuthInfo = attr.ib()
    tx_body: TxBody = attr.ib()

    def to_amino(self) -> dict:
        tx = self.tx_body
        auth = self.auth_info
        return {
            "chain_id": self.chain_id,
            "account_number": str(self.account_number),
            "sequence": str(self.sequence),
            "timeout_height": str(tx.timeout_height)
            if (tx.timeout_height is not None and tx.timeout_height != 0)
            else None,
            "fee": auth.fee.to_amino(),
            "msgs": [msg.to_amino() for msg in tx.messages],
            "memo": tx.memo if tx.memo else "",
        }

    @classmethod
    def from_data(cls, data: dict) -> SignDoc:
        return cls(
            chain_id=data["chain_id"],
            account_number=data["account_number"],
            sequence=data["sequence"],
            auth_info=AuthInfo.from_data(data["auth_info"]),
            tx_body=TxBody.from_data(data["tx_body"]),
        )

    def to_data(self) -> dict:
        return {
            "chain_id": self.chain_id,
            "account_nubmer": self.account_number,
            "sequence": self.sequence,
            "auth_info": self.auth_info.to_data(),
            "tx_body": self.tx_body.to_data(),
        }

    @classmethod
    def from_proto(cls, proto: SignDoc_pb) -> SignDoc:
        return cls(
            chain_id=proto.chain_id,
            account_number=proto.account_number,
            auth_info=AuthInfo.from_proto(proto.auth_info_bytes),
            tx_body=TxBody.from_proto(proto.body_bytes),
        )

    def to_proto(self) -> SignDoc_pb:
        return SignDoc_pb(
            body_bytes=bytes(self.tx_body.to_proto()),
            auth_info_bytes=bytes(self.auth_info.to_proto()),
            chain_id=self.chain_id,
            account_number=self.account_number,
        )

    def to_bytes(self) -> bytes:
        return bytes(self.to_proto())

    def to_amino_json(self) -> bytes:
        amino = self.to_amino()
        return bytes(
            json.dumps(remove_none(amino), sort_keys=True, separators=(",", ":")),
            "utf-8",
        )



import base64
import hashlib


def hash_amino(txdata: str) -> str:
    """Get the transaction hash from Amino-encoded Transaction in base64."""
    return hashlib.sha256(base64.b64decode(txdata)).digest().hex()



from __future__ import annotations

import attr
from terra_proto.cosmos.gov.v1beta1 import Deposit as Deposit_pb

from terra_sdk.core import AccAddress
from terra_sdk.util.json import JSONSerializable

from .coins import Coins


@attr.s
class Deposit(JSONSerializable):
    proposal_id: int = attr.ib(converter=int)
    depositor: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    @classmethod
    def from_data(cls, data: dict) -> Deposit:
        return cls(
            proposal_id=data.get("proposal_id"),
            depositor=data.get("depositor"),
            amount=Coins.from_data(data.get("amount")),
        )

    @classmethod
    def from_proto(cls, proto: Deposit_pb) -> Deposit:
        return cls(
            proposal_id=proto.proposal_id,
            depositor=proto.depositor,
            amount=Coins.from_proto(proto.amount),
        )

    def to_proto(self) -> Deposit_pb:
        return Deposit_pb(
            proposal_id=self.proposal_id,
            depositor=self.depositor,
            amount=self.amount.to_proto(),
        )



"""Some useful base classes to inherit from."""
from abc import abstractmethod
from typing import Any, Callable, Dict, List

import attr
from betterproto import Message
from betterproto.lib.google.protobuf import Any as Any_pb

from .json import JSONSerializable, dict_to_data


class BaseTerraData(JSONSerializable, Message):

    type: str
    type_url: str

    def to_data(self) -> dict:
        data = dict_to_data(attr.asdict(self))
        data.update({"@type": self.type_url})
        return data

    @abstractmethod
    def to_proto(self):
        pass


# data demux
def create_demux(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_url: i.from_data for i in inputs}

    def from_data(data: dict):
        return table[data["@type"]](data)

    return from_data


# for other protos inside of msgs
def create_demux_proto(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_url: i.from_proto for i in inputs}

    def from_proto(proto: Any_pb):
        return table[proto.type_url](proto)

    return from_proto


# Any_pb to Proto for msgs
def create_demux_unpack_any(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_url: i.from_proto for i in inputs}
    prototypes = {i.type_url: i.prototype for i in inputs}

    def unpack_any(proto: Any_pb):
        return table[proto.type_url](prototypes[proto.type_url]().parse(proto.value))

    return unpack_any


# legacy amino demux
def create_demux_amino(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_amino: i.from_amino for i in inputs}

    def from_amino(data: dict):
        return table[data["type"]](data)

    return from_amino



from .channel import (
    MsgAcknowledgement,
    MsgChannelCloseConfirm,
    MsgChannelCloseInit,
    MsgChannelOpenAck,
    MsgChannelOpenConfirm,
    MsgChannelOpenInit,
    MsgChannelOpenTry,
    MsgRecvPacket,
    MsgTimeout,
)
from .client import (
    MsgCreateClient,
    MsgSubmitMisbehaviour,
    MsgUpdateClient,
    MsgUpgradeClient,
)
from .connection import (
    MsgConnectionOpenAck,
    MsgConnectionOpenConfirm,
    MsgConnectionOpenInit,
    MsgConnectionOpenTry,
)

__all__ = [
    "MsgCreateClient",
    "MsgUpdateClient",
    "MsgUpgradeClient",
    "MsgSubmitMisbehaviour",
    "MsgConnectionOpenInit",
    "MsgConnectionOpenTry",
    "MsgConnectionOpenAck",
    "MsgConnectionOpenConfirm",
    "MsgChannelOpenInit",
    "MsgChannelOpenTry",
    "MsgChannelOpenAck",
    "MsgChannelOpenConfirm",
    "MsgChannelCloseInit",
    "MsgChannelCloseConfirm",
    "MsgRecvPacket",
    "MsgTimeout",
    "MsgAcknowledgement",
]



from terra_sdk.core import Dec, Numeric

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncMintAPI", "MintAPI"]


class AsyncMintAPI(BaseAsyncAPI):
    async def inflation(self) -> Dec:
        """Fetches the current inflation.

        Returns:
            Dec: inflation
        """
        res = await self._c._get("/cosmos/mint/v1beta1/inflation")
        return Dec(res.get("inflation"))

    async def annual_provisions(self) -> Dec:
        """Fetches the annual provisions.

        Returns:
            Dec: annual provisions
        """
        res = await self._c._get("/cosmos/mint/v1beta1/annual_provisions")
        return Dec(res.get("annual_provisions"))

    async def parameters(self) -> dict:
        """Fetches the Mint module's parameters.

        Returns:
            dict: Mint module parameters
        """
        res = await self._c._get("/cosmos/mint/v1beta1/params")
        params = res.get("params")
        return {
            "mint_denom": params["mint_denom"],
            "inflation_rate_change": Dec(params["inflation_rate_change"]),
            "inflation_max": Dec(params["inflation_max"]),
            "inflation_min": Dec(params["inflation_min"]),
            "goal_bonded": Dec(params["goal_bonded"]),
            "blocks_per_year": Numeric.parse(params["blocks_per_year"]),
        }


class MintAPI(AsyncMintAPI):
    @sync_bind(AsyncMintAPI.inflation)
    def inflation(self) -> Dec:
        pass

    inflation.__doc__ = AsyncMintAPI.inflation.__doc__

    @sync_bind(AsyncMintAPI.annual_provisions)
    def annual_provisions(self) -> Dec:
        pass

    annual_provisions.__doc__ = AsyncMintAPI.annual_provisions.__doc__

    @sync_bind(AsyncMintAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncMintAPI.parameters.__doc__



from typing import Optional

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncTendermintAPI", "TendermintAPI"]


class AsyncTendermintAPI(BaseAsyncAPI):
    async def node_info(self) -> dict:
        """Fetches the curent connected node's information.

        Returns:
            dict: node information
        """
        res = await self._c._get("/cosmos/base/tendermint/v1beta1/node_info")
        return {
            "default_node_info": res["default_node_info"],
            "application_version": res["application_version" ""],
        }

    async def syncing(self) -> bool:
        """Fetches whether the curent connect node is syncing with the network.

        Returns:
            bool: syncing status
        """
        return (await self._c._get("/cosmos/base/tendermint/v1beta1/syncing"))[
            "syncing"
        ]

    async def validator_set(self, height: Optional[int] = None) -> dict:
        """Fetches the validator set for a height. If no height is given, defaults to latest.

        Args:
            height (int, optional): block height.

        Returns:
            dict: validator set
        """
        x = "latest" if height is None else height
        return await self._c._get(f"/cosmos/base/tendermint/v1beta1/validatorsets/{x}")

    async def block_info(self, height: Optional[int] = None) -> dict:
        """Fetches the block information for a given height. If no height is given, defaults to latest block.

        Args:
            height (int, optional): block height.

        Returns:
            dict: block info
        """
        x = "latest" if height is None else height
        return await self._c._get(f"/cosmos/base/tendermint/v1beta1/blocks/{x}")


class TendermintAPI(AsyncTendermintAPI):
    @sync_bind(AsyncTendermintAPI.node_info)
    def node_info(self) -> dict:
        pass

    node_info.__doc__ = AsyncTendermintAPI.node_info.__doc__

    @sync_bind(AsyncTendermintAPI.syncing)
    def syncing(self) -> bool:
        pass

    syncing.__doc__ = AsyncTendermintAPI.syncing.__doc__

    @sync_bind(AsyncTendermintAPI.validator_set)
    def validator_set(self, height: Optional[int] = None) -> dict:
        pass

    validator_set.__doc__ = AsyncTendermintAPI.validator_set.__doc__

    @sync_bind(AsyncTendermintAPI.block_info)
    def block_info(self, height: Optional[int] = None) -> dict:
        pass

    block_info.__doc__ = AsyncTendermintAPI.block_info.__doc__



"""Data objects pertaining to building, signing, and parsing Transactions."""

from __future__ import annotations

import base64
import json
from typing import Dict, List, Optional

import attr
from betterproto.lib.google.protobuf import Any
from terra_proto.cosmos.base.abci.v1beta1 import AbciMessageLog as AbciMessageLog_pb
from terra_proto.cosmos.base.abci.v1beta1 import Attribute as Attribute_pb
from terra_proto.cosmos.base.abci.v1beta1 import StringEvent as StringEvent_pb
from terra_proto.cosmos.base.abci.v1beta1 import TxResponse as TxResponse_pb
from terra_proto.cosmos.tx.signing.v1beta1 import SignMode as SignMode_pb
from terra_proto.cosmos.tx.v1beta1 import AuthInfo as AuthInfo_pb
from terra_proto.cosmos.tx.v1beta1 import SignerInfo as SignerInfo_pb
from terra_proto.cosmos.tx.v1beta1 import Tx as Tx_pb
from terra_proto.cosmos.tx.v1beta1 import TxBody as TxBody_pb

from terra_sdk.core.compact_bit_array import CompactBitArray
from terra_sdk.core.fee import Fee
from terra_sdk.core.mode_info import ModeInfo, ModeInfoMulti, ModeInfoSingle
from terra_sdk.core.msg import Msg
from terra_sdk.core.public_key import (
    LegacyAminoMultisigPublicKey,
    PublicKey,
    SimplePublicKey,
)
from terra_sdk.core.signature_v2 import SignatureV2
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "SignMode",
    "AuthInfo",
    "Tx",
    "TxBody",
    "TxLog",
    "TxInfo",
    "parse_tx_logs",
    "SignerInfo",
    "SignerData",
]

# just alias
from terra_sdk.util.parse_msg import parse_proto

SignMode = SignMode_pb


@attr.s
class SignerData:
    sequence: int = attr.ib(converter=int)
    public_key: Optional[PublicKey] = attr.ib(default=None)


@attr.s
class Tx(JSONSerializable):
    """Data structure for a transaction which can be broadcasted.

    Args:
        body (TxBody): the processable content of the transaction
        auth_info (AuthInfo): the authorization related content of the transaction
        signatures (List[bytes]): signatures is a list of signatures that matches the length and order of body and auth_info
    """

    body: TxBody = attr.ib()
    auth_info: AuthInfo = attr.ib()
    signatures: List[bytes] = attr.ib(converter=list)

    def to_data(self) -> dict:
        return {
            "body": self.body.to_data(),
            "auth_info": self.auth_info.to_data(),
            "signatures": [
                base64.b64encode(sig).decode("ascii") for sig in self.signatures
            ],
        }

    def to_proto(self) -> Tx_pb:
        return Tx_pb(
            body=self.body.to_proto(),
            auth_info=self.auth_info.to_proto(),
            signatures=self.signatures,
        )

    @classmethod
    def from_data(cls, data: dict) -> Tx:
        return cls(
            TxBody.from_data(data["body"]),
            AuthInfo.from_data(data["auth_info"]),
            [base64.b64decode(sig) for sig in data["signatures"]],
        )

    @classmethod
    def from_proto(cls, proto: Tx_pb) -> Tx:
        return cls(
            TxBody.from_proto(proto.body),
            AuthInfo.from_proto(proto.auth_info),
            proto.signatures,
        )

    @classmethod
    def from_bytes(cls, txb: bytes) -> Tx:
        proto = Tx_pb().parse(txb)
        c = cls.from_proto(proto)
        return c

    def append_empty_signatures(self, signers: List[SignerData]):
        for signer in signers:
            if signer.public_key is not None:
                if isinstance(signer.public_key, LegacyAminoMultisigPublicKey):
                    signer_info = SignerInfo(
                        public_key=signer.public_key,
                        sequence=signer.sequence,
                        mode_info=ModeInfo(
                            multi=ModeInfoMulti(
                                CompactBitArray.from_bits(
                                    len(signer.public_key.public_keys)
                                ),
                                [],
                            )
                        ),
                    )
                else:
                    signer_info = SignerInfo(
                        public_key=signer.public_key,
                        sequence=signer.sequence,
                        mode_info=ModeInfo(
                            ModeInfoSingle(mode=SignMode.SIGN_MODE_DIRECT)
                        ),
                    )
            else:
                signer_info = SignerInfo(
                    public_key=SimplePublicKey(""),
                    sequence=signer.sequence,
                    mode_info=ModeInfo(ModeInfoSingle(mode=SignMode.SIGN_MODE_DIRECT)),
                )
            self.auth_info.signer_infos.append(signer_info)
            self.signatures.append(b" ")

    def clear_signature(self):
        self.signatures.clear()
        self.auth_info.signer_infos.clear()

    def append_signatures(self, signatures: List[SignatureV2]):
        for sig in signatures:
            mode_info, sig_bytes = sig.data.to_mode_info_and_signature()
            self.signatures.append(sig_bytes)
            # self.signatures.append(base64.b64decode(sig_bytes))
            self.auth_info.signer_infos.append(
                SignerInfo(sig.public_key, mode_info, sig.sequence)
            )


@attr.s
class TxBody(JSONSerializable):
    """Body of a transaction.

    Args:
        messages: list of messages to include in transaction
        memo: transaction memo
        timeout_height:
    """

    messages: List[Msg] = attr.ib()
    memo: Optional[str] = attr.ib(default="")
    timeout_height: int = attr.ib(
        default=0, converter=int
    )  # TxBody_pb.timeout_height is int

    def to_data(self) -> dict:
        return {
            "messages": [m.to_data() for m in self.messages],
            "memo": self.memo,
            "timeout_height": self.timeout_height,
        }

    def to_proto(self) -> TxBody_pb:
        return TxBody_pb(
            messages=[m.pack_any() for m in self.messages],
            memo=self.memo,
            timeout_height=self.timeout_height,
        )

    @classmethod
    def from_data(cls, data: dict) -> TxBody:
        return cls(
            [Msg.from_data(m) for m in data["messages"]],
            data["memo"],
            data["timeout_height"] if data["timeout_height"] else 0,
        )

    @classmethod
    def from_proto(cls, proto: TxBody_pb) -> TxBody:
        return cls(
            [Msg.unpack_any(m) for m in proto.messages],
            proto.memo,
            proto.timeout_height,
        )


@attr.s
class AuthInfo(JSONSerializable):
    """AuthInfo

    Args:
        signer_infos: information of the signers
        fee: Fee
    """

    signer_infos: List[SignerInfo] = attr.ib(converter=list)
    fee: Fee = attr.ib()

    def to_dict(self, casing, include_default_values) -> dict:
        return self.to_proto().to_dict(casing, include_default_values)

    def to_data(self) -> dict:
        return {
            "signer_infos": [si.to_data() for si in self.signer_infos],
            "fee": self.fee.to_data(),
        }

    def to_proto(self) -> AuthInfo_pb:
        return AuthInfo_pb(
            signer_infos=[signer.to_proto() for signer in self.signer_infos],
            fee=self.fee.to_proto(),
        )

    @classmethod
    def from_data(cls, data: dict) -> AuthInfo:
        return cls(
            [SignerInfo.from_data(m) for m in data["signer_infos"]],
            Fee.from_data(data["fee"]),
        )

    @classmethod
    def from_proto(cls, proto: AuthInfo_pb) -> AuthInfo:
        return cls(
            [SignerInfo.from_proto(m) for m in proto.signer_infos],
            Fee.from_proto(proto.fee),
        )


@attr.s
class SignerInfo(JSONSerializable):
    """SignerInfo
    Args:
       public_key (PublicKey)
       mode_info (ModeInfo)
       sequence (int)
    """

    public_key: PublicKey = attr.ib()
    mode_info: ModeInfo = attr.ib()
    sequence: int = attr.ib(converter=int)

    def to_data(self) -> dict:
        return {
            "public_key": self.public_key.to_data(),
            "mode_info": self.mode_info.to_data(),
            "sequence": self.sequence,
        }

    def to_proto(self) -> SignerInfo_pb:
        return SignerInfo_pb(
            public_key=self.public_key.pack_any(),
            mode_info=self.mode_info.to_proto(),
            sequence=self.sequence,
        )

    @classmethod
    def from_data(cls, data: dict) -> SignerInfo:
        return cls(
            public_key=PublicKey.from_data(data["public_key"]),
            mode_info=ModeInfo.from_data(data["mode_info"]),
            sequence=data["sequence"],
        )

    @classmethod
    def from_proto(cls, proto: SignerInfo_pb) -> SignerInfo:
        return cls(
            public_key=PublicKey.unpack_any(proto.public_key),
            mode_info=ModeInfo.from_proto(proto.mode_info),
            sequence=proto.sequence,
        )


def parse_events_by_type(event_data: List[dict]) -> Dict[str, Dict[str, List[str]]]:
    events: Dict[str, Dict[str, List[str]]] = {}
    for ev in event_data:
        for att in ev["attributes"]:
            if ev["type"] not in events:
                events[ev["type"]] = {}
            if att["key"] not in events[ev["type"]]:
                events[ev["type"]][att["key"]] = []
            events[ev["type"]][att["key"]].append(att.get("value"))
    return events


@attr.s
class TxLog(JSONSerializable):
    """Object containing the events of a transaction that is automatically generated when
    :class:`TxInfo` or :class:`BlockTxBroadcastResult` objects are read."""

    msg_index: int = attr.ib(converter=int)
    """Number of the message inside the transaction that it was included in."""

    log: str = attr.ib()
    """This field may be populated with details of the message's error, if any."""

    events: List[dict] = attr.ib()
    """Raw event log data"""

    events_by_type: Dict[str, Dict[str, List[str]]] = attr.ib(init=False)
    """Event log data, re-indexed by event type name and attribute type.

    For instance, the event type may be: ``store_code`` and an attribute key could be
    ``code_id``.

    >>> logs[0].events_by_type["<event-type>"]["<attribute-key>"]
    ['<attribute-value>', '<attribute-value2>']
    """

    def __attrs_post_init__(self):
        self.events_by_type = parse_events_by_type(self.events)

    @classmethod
    def from_proto(cls, tx_log: AbciMessageLog_pb) -> TxLog:
        events = [json.loads(event) for event in tx_log.events]
        return cls(msg_index=tx_log.msg_index, log=tx_log.log, events=events)

    def to_proto(self) -> AbciMessageLog_pb:
        str_events = []
        for event in self.events:
            str_events.append(json.dumps(event))
        return AbciMessageLog_pb(
            msg_index=self.msg_index, log=self.log, events=str_events
        )


@attr.s
class Attribute(JSONSerializable):
    key: str = attr.ib()
    value: str = attr.ib()

    def to_proto(self) -> Attribute_pb:
        proto = Attribute_pb()
        proto.key = self.key
        proto.value = self.value
        return proto

    @classmethod
    def from_proto(cls, attrib: Attribute_pb) -> Attribute:
        return cls(key=attrib["key"], value=attrib["value"])


@attr.s
class StringEvent(JSONSerializable):

    type: str = attr.ib()
    attributes = attr.ib()

    def to_proto(self) -> StringEvent_pb:
        return StringEvent_pb(type=self.type, attributes=self.attributes)

    @classmethod
    def from_proto(cls, str_event: StringEvent_pb) -> StringEvent:
        return cls(type=str_event["type"], attributes=str_event["attributes"])


def parse_tx_logs(logs) -> Optional[List[TxLog]]:
    return (
        [
            TxLog(msg_index=i, log=log.get("log"), events=log.get("events"))
            for i, log in enumerate(logs)
        ]
        if logs
        else None
    )


def parse_tx_logs_proto(logs: List[AbciMessageLog_pb]) -> Optional[List[TxLog]]:
    return [TxLog.from_proto(log) for log in logs] if logs else None


@attr.s
class TxInfo(JSONSerializable):
    """Holds information pertaining to a transaction which has been included in a block
    on the blockchain.

    .. note::
        Users are not expected to create this object directly. It is returned by
        :meth:`TxAPI.tx_info()<terra_sdk.client.lcd.api.tx.TxAPI.tx_info>`
    """

    height: int = attr.ib(converter=int)
    """Block height at which transaction was included."""

    txhash: str = attr.ib()
    """Transaction hash."""

    rawlog: str = attr.ib()
    """Event log information as a raw JSON-string."""

    logs: Optional[List[TxLog]] = attr.ib()
    """Event log information."""

    gas_wanted: int = attr.ib(converter=int)
    """Gas requested by transaction."""

    gas_used: int = attr.ib(converter=int)
    """Actual gas amount used."""

    tx: Tx = attr.ib()
    """Transaction object."""

    timestamp: str = attr.ib()
    """Time at which transaction was included."""

    code: Optional[int] = attr.ib(default=None)
    """If this field is not ``None``, the transaction failed at ``DeliverTx`` stage."""

    codespace: Optional[str] = attr.ib(default=None)
    """Error subspace (used alongside ``code``)."""

    def to_data(self) -> dict:
        data = {
            "height": str(self.height),
            "txhash": self.txhash,
            "raw_log": self.rawlog,
            "logs": [log.to_data() for log in self.logs] if self.logs else None,
            "gas_wanted": str(self.gas_wanted),
            "gas_used": str(self.gas_used),
            "timestamp": self.timestamp,
            "tx": self.tx.to_data(),
            "code": self.code,
            "codespace": self.codespace,
        }

        return data

    @classmethod
    def from_data(cls, data: dict) -> TxInfo:
        return cls(
            data.get("height"),
            data.get("txhash"),
            data.get("raw_log"),
            parse_tx_logs(data.get("logs")),
            data.get("gas_wanted"),
            data.get("gas_used"),
            Tx.from_data(data.get("tx")),
            data.get("timestamp"),
            data.get("code"),
            data.get("codespace"),
        )

    def to_proto(self) -> TxResponse_pb:
        proto = TxResponse_pb()
        proto.height = self.height
        proto.txhash = self.txhash
        proto.raw_log = self.rawlog
        proto.logs = [log.to_proto() for log in self.logs] if self.logs else None
        proto.gas_wanted = self.gas_wanted
        proto.gas_used = self.gas_used
        proto.timestamp = self.timestamp
        proto.tx = self.tx.to_proto()
        proto.code = self.code
        proto.codespace = self.codespace
        return proto

    @classmethod
    def from_proto(cls, proto: TxResponse_pb) -> TxInfo:
        return cls(
            height=proto.height,
            txhash=proto.txhash,
            rawlog=proto.raw_log,
            logs=parse_tx_logs_proto(proto.logs),
            gas_wanted=proto.gas_wanted,
            gas_used=proto.gas_used,
            timestamp=proto.timestamp,
            tx=Tx.from_proto(proto.tx),
            code=proto.code,
            codespace=proto.codespace,
        )



"""ibc client module data objects."""
from __future__ import annotations

from typing import List

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.ibc.core.client.v1 import (
    ClientConsensusStates as ClientConsensusStates_pb,
)
from terra_proto.ibc.core.client.v1 import (
    ConsensusStateWithHeight as ConsensusStateWithHeight_pb,
)
from terra_proto.ibc.core.client.v1 import Height as Height_pb
from terra_proto.ibc.core.client.v1 import (
    IdentifiedClientState as IdentifiedClientState_pb,
)
from terra_proto.ibc.core.client.v1 import Params as Params_pb

from terra_sdk.util.json import JSONSerializable

__all__ = [
    "Height",
    "IdentifiedClientState",
    "ConsensusStateWithHeight",
    "ClientConsensusStates",
    "Params",
]


@attr.s
class Height(JSONSerializable):
    revision_number: int = attr.ib(converter=int)
    revision_height: int = attr.ib(converter=int)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Height:
        return cls(
            revision_number=data["revision_number"],
            revision_height=data["revision_height"],
        )

    def to_proto(self) -> Height_pb:
        return Height_pb(
            revision_number=self.revision_number, revision_height=self.revision_height
        )

    @classmethod
    def from_proto(cls, proto: Height_pb) -> Height:
        return cls(
            revision_number=proto.revision_number,
            revision_height=proto.revision_height,
        )


@attr.s
class IdentifiedClientState(JSONSerializable):
    """
    IdentifiedClientState defines a client state with an additional client identifier field.
    """

    client_id: str = attr.ib()
    client_state: dict = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> IdentifiedClientState:
        return cls(
            client_id=data["client_id"],
            client_state=Any_pb().from_dict(data["client_state"]),
        )

    def to_proto(self) -> IdentifiedClientState_pb:
        return IdentifiedClientState_pb(
            client_id=self.client_id,
            client_state=Any_pb().from_dict(self.client_state),
        )

    @classmethod
    def from_proto(cls, proto: IdentifiedClientState_pb) -> IdentifiedClientState:
        return cls(
            client_id=proto.client_id, client_state=proto.client_state.to_dict()
        )


@attr.s
class ConsensusStateWithHeight(JSONSerializable):
    """
    ConsensusStateWithHeight defines a consensus state with an additional height field.
    """

    height: Height = attr.ib()
    consensus_state: dict = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> ConsensusStateWithHeight:
        return cls(
            height=data["height"],
            consensus_state=Any_pb().from_dict(data["consensus_state"]),
        )

    def to_proto(self) -> ConsensusStateWithHeight_pb:
        return ConsensusStateWithHeight_pb(
            height=self.height.to_proto(),
            consensus_state=Any_pb().from_dict(self.consensus_state),
        )

    @classmethod
    def from_proto(cls, proto: ConsensusStateWithHeight_pb) -> ConsensusStateWithHeight:
        return cls(
            height=Height.from_proto(proto.height),
            consensus_state=proto.consensus_state.to_dict(),
        )


@attr.s
class ClientConsensusStates(JSONSerializable):
    """
    ClientConsensusStates defines all the stored consensus states for a given client.
    """

    client_id: str = attr.ib()
    consensus_states: List[ConsensusStateWithHeight] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> ClientConsensusStates:
        return cls(
            client_id=data["client_id"],
            consensus_states=[
                ConsensusStateWithHeight.from_data(state)
                for state in data["consensus_states"]
            ],
        )

    def to_proto(self) -> ClientConsensusStates_pb:
        return ClientConsensusStates_pb(
            client_id=self.client_id,
            consensus_states=[state.to_proto() for state in self.consensus_states],
        )

    @classmethod
    def from_proto(cls, proto: ClientConsensusStates_pb) -> ClientConsensusStates:
        return cls(
            client_id=proto.client_id,
            consensus_states=[
                ConsensusStateWithHeight.from_proto(state)
                for state in proto.consensus_states
            ],
        )


@attr.s
class Params(JSONSerializable):
    """
    Params defines the set of IBC light client parameters.
    """

    allowed_clients: List[str] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Params:
        return cls(allowed_clients=data["allowed_clients"])

    def to_proto(self) -> Params_pb:
        return Params_pb(allowed_clients=self.allowed_clients)

    @classmethod
    def from_proto(cls, proto: Params_pb) -> Params:
        return cls(allowed_clients=proto.allowed_clients)



from typing import Optional

from terra_sdk.core import AccAddress
from terra_sdk.core.feegrant import Allowance

from ..params import APIParams
from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncFeeGrantAPI", "FeeGrantAPI"]


class AsyncFeeGrantAPI(BaseAsyncAPI):
    async def allowances(
        self, address: AccAddress, params: Optional[APIParams] = None
    ) -> (Allowance, dict):
        """fetch fee allowances

        Args:
            address (AccAddress): grantee address
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            Allowances[]: granted allowances
            pagination[]: pagination info
        """
        res = await self._c._get(
            f"cosmos/feegrant/v1beta1/allowances/{address}", params
        )
        allowances = []
        for i in res["allowances"]:
            allowance = {
                "granter": i.get("granter"),
                "grantee": i.get("grantee"),
                "allowance": Allowance.from_data(i.get("allowance")),
            }
            allowances.append(allowance)
        return allowances, res.get("pagination")

    async def allowance(
        self,
        granter: AccAddress,
        grantee: AccAddress,
        params: Optional[APIParams] = None,
    ) -> Allowance:
        """fetch granter's allowance for the grantee

        Args:
            granter (AccAddress): granter is the address of the user granting an allowance of their funds.
            grantee (AccAddress): grantee is the address of the user being granted an allowance of another user’s funds.
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            Allowance: granted allowance
        """
        res = await self._c._get(
            f"cosmos/feegrant/v1beta1/allowance/{granter}/{grantee}", params
        )
        res = res.get("allowance")
        return {
            "granter": res.get("granter"),
            "grantee": res.get("grantee"),
            "allowance": Allowance.from_data(res.get("allowance")),
        }


class FeeGrantAPI(AsyncFeeGrantAPI):
    @sync_bind(AsyncFeeGrantAPI.allowances)
    def allowances(
        self, address: AccAddress, params: Optional[APIParams] = None
    ) -> (Allowance, dict):
        pass

    allowances.__doc__ = AsyncFeeGrantAPI.allowances.__doc__

    @sync_bind(AsyncFeeGrantAPI.allowance)
    def allowance(self) -> Allowance:
        pass

    allowance.__doc__ = AsyncFeeGrantAPI.allowance.__doc__



"""Gov module message types."""

from __future__ import annotations

import attr
from terra_proto.cosmos.gov.v1beta1 import MsgDeposit as MsgDeposit_pb
from terra_proto.cosmos.gov.v1beta1 import MsgSubmitProposal as MsgSubmitProposal_pb
from terra_proto.cosmos.gov.v1beta1 import MsgVote as MsgVote_pb

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg

from .data import Content, VoteOption

__all__ = ["MsgSubmitProposal", "MsgDeposit", "MsgVote"]


@attr.s
class MsgSubmitProposal(Msg):
    """Submit the attached proposal with an initial deposit.

    Args:
        content (Content): type of proposal
        initial_deposit (Coins): initial deposit for proposal made by proposer
        proposer (AccAddress): proposal submitter
    """

    type_amino = "cosmos-sdk/MsgSubmitProposal"
    """"""
    type_url = "/cosmos.gov.v1beta1.MsgSubmitProposal"
    """"""
    action = "submit_proposal"
    """"""
    prototype = MsgSubmitProposal_pb
    """"""

    content: Content = attr.ib()
    initial_deposit: Coins = attr.ib(converter=Coins)
    proposer: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "content": self.content.to_amino(),
                "initial_deposit": self.initial_deposit.to_amino(),
                "proposer": self.proposer,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "content": self.content.to_data(),
            "initial_deposit": self.initial_deposit.to_data(),
            "proposer": self.proposer,
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitProposal:
        from terra_sdk.util.parse_content import parse_content

        content = parse_content(data["content"])
        return cls(
            content=content,
            initial_deposit=Coins.from_data(data["initial_deposit"]),
            proposer=data["proposer"],
        )

    def to_proto(self) -> MsgSubmitProposal_pb:
        return MsgSubmitProposal_pb(
            content=self.content.to_proto(),
            initial_deposit=self.initial_deposit.to_proto(),
            proposer=self.proposer,
        )

    @classmethod
    def from_proto(cls, proto: MsgSubmitProposal_pb) -> MsgSubmitProposal:
        from terra_sdk.util.parse_content import parse_content_unpack_any

        content = parse_content_unpack_any(proto.content)
        return cls(
            content=content,
            initial_deposit=Coins.from_proto(proto.initial_deposit),
            proposer=proto.proposer,
        )


@attr.s
class MsgDeposit(Msg):
    """Deposit funds for an active deposit-stage proposal.

    Args:
        proposal_id (int): proposal number to deposit for
        depositor (AccAddress): account making deposit
        amount (Coins): amount to deposit
    """

    type_amino = "cosmos-sdk/MsgDeposit"
    """"""
    type_url = "/cosmos.gov.v1beta1.MsgDeposit"
    """"""
    action = "deposit"
    """"""
    prototype = MsgDeposit_pb
    """"""

    proposal_id: int = attr.ib(converter=int)
    depositor: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "proposal_id": str(self.proposal_id),
                "depositor": self.depositor,
                "amount": self.amount.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "proposal_id": str(self.proposal_id),
            "depositor": self.depositor,
            "amount": self.amount.to_data(),
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgDeposit:
        return cls(
            proposal_id=data["proposal_id"],
            depositor=data["depositor"],
            amount=Coins.from_data(data["amount"]),
        )

    def to_proto(self) -> MsgDeposit_pb:
        return MsgDeposit_pb(
            proposal_id=self.proposal_id,
            depositor=self.depositor,
            amount=self.amount.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgDeposit_pb) -> MsgDeposit:
        return cls(
            proposal_id=proto.proposal_id,
            depositor=proto.depositor,
            amount=Coins.from_proto(proto.amount),
        )


@attr.s
class MsgVote(Msg):
    """Vote for an active voting-stage proposal.

    Args:
        proposal_id (int): proposal to vote for
        voter (AccAddress): account casting vote
        option (VoteOption): vote option (must be one of: :data:`MsgVote.ABSTAIN`, :data:`MsgVote.YES`, :data:`MsgVote.NO`, or :data:`MsgVote.NO_WITH_VETO`),
    """

    type_amino = "cosmos-sdk/MsgVote"
    """"""
    type_url = "/cosmos.gov.v1beta1.MsgVote"
    """"""
    action = "vote"
    """"""
    prototype = MsgVote_pb
    """"""

    EMPTY = "Empty"
    """Encodes an empty vote option."""

    YES = "Yes"
    """"""
    ABSTAIN = "Abstain"
    """"""
    NO = "No"
    """"""
    NO_WITH_VETO = "NoWithVeto"
    """"""

    proposal_id: int = attr.ib(converter=int)
    voter: AccAddress = attr.ib()
    option: VoteOption = attr.ib()

    """
    @option.validator
    def _check_option(self, attribute, value):
        possible_options = [
            self.EMPTY,
            self.YES,
            self.ABSTAIN,
            self.NO,
            self.NO_WITH_VETO,
        ]
        if value not in possible_options:
            raise TypeError(
                f"incorrect value for option: {value}, must be one of: {possible_options}"
            )
    """

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "proposal_id": str(self.proposal_id),
                "voter": self.voter,
                "option": self.option.name,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgVote:
        return cls(
            proposal_id=data["proposal_id"],
            voter=data["voter"],
            option=data["option"],
        )

    def to_proto(self) -> MsgVote_pb:
        return MsgVote_pb(
            proposal_id=self.proposal_id, voter=self.voter, options=self.option
        )

    @classmethod
    def from_proto(cls, proto: MsgVote_pb) -> MsgVote:
        return cls(
            proposal_id=proto.proposal_id, voter=proto.voter, option=proto.option
        )



from typing import Union

from terra_proto.cosmos.distribution.v1beta1 import (
    CommunityPoolSpendProposal as CommunityPoolSpendProposal_pb,
)
from terra_proto.cosmos.gov.v1beta1 import TextProposal as TextProposal_pb
from terra_proto.cosmos.params.v1beta1 import (
    ParameterChangeProposal as ParameterChangeProposal_pb,
)
from terra_proto.cosmos.upgrade.v1beta1 import (
    CancelSoftwareUpgradeProposal as CancelSoftwareUpgradeProposal_pb,
)
from terra_proto.cosmos.upgrade.v1beta1 import (
    SoftwareUpgradeProposal as SoftwareUpgradeProposal_pb,
)
from terra_proto.ibc.core.client.v1 import (
    ClientUpdateProposal as ClientUpdateProposal_pb,
)

from terra_sdk.core.distribution.proposals import CommunityPoolSpendProposal
from terra_sdk.core.gov.proposals import TextProposal
from terra_sdk.core.ibc.proposals import ClientUpdateProposal
from terra_sdk.core.params.proposals import ParameterChangeProposal
from terra_sdk.core.upgrade import (
    CancelSoftwareUpgradeProposal,
    SoftwareUpgradeProposal,
)

from .base import create_demux, create_demux_proto, create_demux_unpack_any

Content = Union[
    TextProposal,
    CommunityPoolSpendProposal,
    ParameterChangeProposal,
    SoftwareUpgradeProposal,
    CancelSoftwareUpgradeProposal,
    ClientUpdateProposal,
]

parse_content = create_demux(
    [
        CommunityPoolSpendProposal,
        TextProposal,
        ParameterChangeProposal,
        SoftwareUpgradeProposal,
        CancelSoftwareUpgradeProposal,
        ClientUpdateProposal,
    ]
)

parse_content_proto = create_demux_proto(
    [
        CommunityPoolSpendProposal,
        TextProposal,
        ParameterChangeProposal,
        SoftwareUpgradeProposal,
        CancelSoftwareUpgradeProposal,
        ClientUpdateProposal,
    ]
)

parse_content_unpack_any = create_demux_unpack_any(
    [
        CommunityPoolSpendProposal,
        TextProposal,
        ParameterChangeProposal,
        SoftwareUpgradeProposal,
        CancelSoftwareUpgradeProposal,
        ClientUpdateProposal,
    ]
)

"""
parse_content_proto = create_demux_proto(
    [
        [CommunityPoolSpendProposal.type_url, CommunityPoolSpendProposal_pb],
        [TextProposal.type_url, TextProposal_pb],
        [ParameterChangeProposal.type_url, ParameterChangeProposal_pb],
        [SoftwareUpgradeProposal.type_url, SoftwareUpgradeProposal_pb],
        [CancelSoftwareUpgradeProposal.type_url, CancelSoftwareUpgradeProposal_pb],
        [ClientUpdateProposal.type_url, ClientUpdateProposal_pb]
    ]
)
"""



"""Authz module message types."""

from __future__ import annotations

from typing import List

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.authz.v1beta1 import MsgExec as MsgExec_pb
from terra_proto.cosmos.authz.v1beta1 import MsgGrant as MsgGrant_pb
from terra_proto.cosmos.authz.v1beta1 import MsgRevoke as MsgRevoke_pb

from terra_sdk.core import AccAddress
from terra_sdk.core.msg import Msg

from .data import Authorization, AuthorizationGrant

__all__ = ["MsgExecAuthorized", "MsgGrantAuthorization", "MsgRevokeAuthorization"]


@attr.s
class MsgExecAuthorized(Msg):
    """Execute a set of messages, exercising an existing authorization.

    Args:
        grantee: grantee account (submitting on behalf of granter)
        msg (List[Msg]): list of messages to execute using authorization grant
    """

    type_amino = "cosmos-sdk/MsgExecAuthorized"
    """"""
    type_url = "/cosmos.authz.v1beta1.MsgExec"
    """"""
    prototype = MsgExec_pb
    """"""

    grantee: AccAddress = attr.ib()
    msgs: List[Msg] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "grantee": self.grantee,
                "msgs": [msg.to_amino() for msg in self.msgs],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "grantee": self.grantee,
            "msgs": [msg.to_data() for msg in self.msgs],
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgExecAuthorized:
        return cls(
            grantee=data["grantee"], msgs=[Msg.from_data(md) for md in data["msgs"]]
        )

    def to_proto(self) -> MsgExec_pb:
        return MsgExec_pb(grantee=self.grantee, msgs=[m.pack_any() for m in self.msgs])

    @classmethod
    def from_proto(cls, proto: MsgExec_pb) -> MsgExecAuthorized:
        return cls(
            grantee=proto.grantee, msgs=[Msg.from_proto(md) for md in proto.msgs]
        )

    @classmethod
    def from_amino(cls, amino: dict) -> MsgExecAuthorized:
        value = amino["value"]
        return cls(
            grantee=value["grantee"],
            msgs=[Msg.from_amino(msg) for msg in value["msgs"]],
        )

    @classmethod
    def unpack_any(cls, any_pb: Any_pb) -> MsgExecAuthorized:
        return cls.from_proto(MsgExec_pb().parse(any_pb.value))


@attr.s
class MsgGrantAuthorization(Msg):
    """Grant an authorization to ``grantee`` to call messages on behalf of ``granter``.

    Args:
        granter: account granting authorization
        grantee: account receiving authorization
        grant: pair of authorization, expiration
    """

    type_amino = "cosmos-sdk/MsgGrantAuthorization"
    """"""
    type_url = "/cosmos.authz.v1beta1.MsgGrant"
    """"""
    prototype = MsgGrant_pb
    """"""

    granter: AccAddress = attr.ib()
    grantee: AccAddress = attr.ib()
    grant: AuthorizationGrant = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "granter": self.granter,
                "grantee": self.grantee,
                "grant": self.grant.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "granter": self.granter,
            "grantee": self.grantee,
            "grant": self.grant.to_data(),
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgGrantAuthorization:
        return cls(
            granter=data["granter"],
            grantee=data["grantee"],
            grant=AuthorizationGrant.from_data(data["grant"]),
        )

    def to_proto(self) -> MsgGrant_pb:
        return MsgGrant_pb(
            granter=self.granter, grantee=self.grantee, grant=self.grant.to_proto()
        )

    @classmethod
    def from_proto(cls, proto: MsgGrant_pb) -> MsgGrantAuthorization:
        return cls(
            granter=proto.granter,
            grantee=proto.grantee,
            grant=AuthorizationGrant.from_proto(proto.grant),
        )

    @classmethod
    def from_amino(cls, amino: dict) -> MsgGrantAuthorization:
        value = amino["value"]
        return cls(
            grantee=value["grantee"],
            granter=value["granter"],
            grant=AuthorizationGrant.from_amino(value["grant"]),
        )


@attr.s
class MsgRevokeAuthorization(Msg):
    """Remove existing authorization grant of the specified message type.

    Args:
        granter: account removing authorization
        grantee: account having authorization removed
        msg_type_url: type of message to remove authorization for
    """

    type_amino = "cosmos-sdk/MsgRevokeAuthorization"
    """"""
    type_url = "/cosmos.authz.v1beta1.MsgRevoke"
    """"""
    prototype = MsgRevoke_pb
    """"""

    granter: AccAddress = attr.ib()
    grantee: AccAddress = attr.ib()
    msg_type_url: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "granter": self.granter,
                "grantee": self.grantee,
                "msg_type_url": self.msg_type_url,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "granter": self.granter,
            "grantee": self.grantee,
            "msg_type_url": self.msg_type_url,
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgRevokeAuthorization:
        return cls(
            granter=data["granter"],
            grantee=data["grantee"],
            msg_type_url=data["msg_type_url"],
        )

    def to_proto(self) -> MsgRevoke_pb:
        return MsgRevoke_pb(
            granter=self.granter, grantee=self.grantee, msg_type_url=self.msg_type_url
        )

    @classmethod
    def from_proto(cls, proto: MsgRevoke_pb) -> MsgRevokeAuthorization:
        return cls(
            granter=proto.granter,
            grantee=proto.grantee,
            msg_type_url=proto.msg_type_url,
        )



"""ibc connection module data objects."""
from __future__ import annotations

from typing import List

import attr
from terra_proto.ibc.core.connection.v1 import Counterparty as Counterparty_pb
from terra_proto.ibc.core.connection.v1 import Version as Version_pb

from terra_sdk.util.json import JSONSerializable

from .commitment import MerklePrefix

__all__ = ["Version", "Counterparty"]


@attr.s
class Version(JSONSerializable):
    identifier: str = attr.ib()
    features: List[str] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Version:
        return cls(
            identifier=data["identifier"],
            features=data["features"],
        )

    def to_proto(self) -> Version_pb:
        return Version_pb(identifier=self.identifier, features=self.features)

    @classmethod
    def from_proto(cls, proto: Version_pb) -> Version:
        return cls(identifier=proto.identifier, features=proto.features)


@attr.s
class Counterparty(JSONSerializable):
    client_id: str = attr.ib()
    connection_id: str = attr.ib()
    prefix: MerklePrefix = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Counterparty:
        return cls(
            client_id=data["client_id"],
            connection_id=data["client_id"],
            prefix=data["prefix"],
        )

    def to_proto(self) -> Counterparty_pb:
        return Counterparty_pb(
            client_id=self.client_id,
            connection_id=self.connection_id,
            prefix=self.prefix.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: Counterparty_pb) -> Counterparty:
        return cls(
            client_id=proto.client_id,
            connection_id=proto.connection_id,
            prefix=MerklePrefix.from_proto(proto.prefix),
        )



from .data import Allowance, AllowedMsgAllowance, BasicAllowance, PeriodicAllowance
from .msgs import MsgGrantAllowance, MsgRevokeAllowance

__all__ = [
    "BasicAllowance",
    "PeriodicAllowance",
    "Allowance",
    "AllowedMsgAllowance",
    "MsgGrantAllowance",
    "MsgRevokeAllowance",
]



__all__ = [
    "Coin",
    "Coins",
    "Dec",
    "Numeric",
    "PublicKey",
    "AccAddress",
    "AccPubKey",
    "ValAddress",
    "SimplePublicKey",
    "LegacyAminoMultisigPublicKey",
    "ValConsPubKey",
    "ValPubKey",
    "SignDoc",
    "CompactBitArray",
    "SignatureV2",
    "MultiSignature",
    "Tx",
    "TxInfo",
    "TxLog",
    "ModeInfo",
    "ModeInfoSingle",
    "ModeInfoMulti",
]

from .bech32 import AccAddress, AccPubKey, ValAddress, ValPubKey
from .coin import Coin
from .coins import Coins
from .compact_bit_array import CompactBitArray
from .multisig import MultiSignature
from .numeric import Dec, Numeric
from .public_key import (
    LegacyAminoMultisigPublicKey,
    PublicKey,
    SimplePublicKey,
    ValConsPubKey,
)
from .sign_doc import SignDoc
from .signature_v2 import SignatureV2
from .tx import ModeInfo, ModeInfoMulti, ModeInfoSingle, Tx, TxInfo, TxLog



"""Upgrade module data objects."""

from __future__ import annotations

__all__ = ["SoftwareUpgradeProposal", "CancelSoftwareUpgradeProposal"]

from typing import Optional

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.upgrade.v1beta1 import (
    CancelSoftwareUpgradeProposal as CancelSoftwareUpgradeProposal_pb,
)
from terra_proto.cosmos.upgrade.v1beta1 import (
    SoftwareUpgradeProposal as SoftwareUpgradeProposal_pb,
)

from terra_sdk.core.upgrade.plan import Plan
from terra_sdk.util.json import JSONSerializable


@attr.s
class SoftwareUpgradeProposal(JSONSerializable):
    title: str = attr.ib()
    description: str = attr.ib()
    plan: Optional[Plan] = attr.ib()

    type_amino = "cosmos-sdk/SoftwareUpgradeProposal"
    """"""
    type_url = "/cosmos.upgrade.v1beta1.SoftwareUpgradeProposal"
    """"""
    prototype = SoftwareUpgradeProposal_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "plan": self.plan.to_amino() if self.plan else None,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> SoftwareUpgradeProposal:
        return cls(
            title=data["title"],
            description=data["description"],
            plan=Plan.from_data(data["plan"]) if data.get("plan") else None,
        )

    def to_proto(self) -> SoftwareUpgradeProposal_pb:
        return SoftwareUpgradeProposal_pb(
            title=self.title,
            description=self.description,
            plan=(self.plan.to_proto() if self.plan else None),
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_proto(cls, proto: SoftwareUpgradeProposal_pb) -> SoftwareUpgradeProposal:
        return cls(
            title=proto.title,
            description=proto.description,
            plan=Plan.from_proto(proto.plan) if proto.plan else None,
        )


@attr.s
class CancelSoftwareUpgradeProposal(JSONSerializable):
    title: str = attr.ib()
    description: str = attr.ib()

    type_amino = "cosmos-sdk/CancelSoftwareUpgradeProposal"
    """"""
    type_url = "/cosmos.upgrade.v1beta1.CancelSoftwareUpgradeProposal"
    """"""
    prototype = CancelSoftwareUpgradeProposal_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> CancelSoftwareUpgradeProposal:
        return cls(title=data["title"], description=data["description"])

    def to_proto(self) -> CancelSoftwareUpgradeProposal_pb:
        return CancelSoftwareUpgradeProposal_pb(
            title=self.title, description=self.description
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_proto(
        cls, proto: CancelSoftwareUpgradeProposal_pb
    ) -> CancelSoftwareUpgradeProposal:
        return cls(title=proto.title, description=proto.description)



"""Staking module message types."""

from __future__ import annotations

from typing import Optional

import attr
from terra_proto.cosmos.staking.v1beta1 import (
    MsgBeginRedelegate as MsgBeginRedelegate_pb,
)
from terra_proto.cosmos.staking.v1beta1 import (
    MsgCreateValidator as MsgCreateValidator_pb,
)
from terra_proto.cosmos.staking.v1beta1 import MsgDelegate as MsgDelegate_pb
from terra_proto.cosmos.staking.v1beta1 import MsgEditValidator as MsgEditValidator_pb
from terra_proto.cosmos.staking.v1beta1 import MsgUndelegate as MsgUndelegate_pb

from terra_sdk.core import AccAddress, Coin, Dec, PublicKey, ValAddress, ValConsPubKey
from terra_sdk.core.msg import Msg

from .data import CommissionRates, Description

__all__ = [
    "MsgBeginRedelegate",
    "MsgDelegate",
    "MsgUndelegate",
    "MsgEditValidator",
    "MsgCreateValidator",
]


@attr.s
class MsgBeginRedelegate(Msg):
    """Redelegate staked Luna from ``validator_src_address`` to ``valdiator_dst_address``.

    Args:
        delegator_address: delegator
        validator_src_address: validator to remove delegation FROM
        validator_dst_address: validator to transfer delegate TO
        amount (Union[str, dict, Coin]): coin (LUNA) to redelegate
    """

    type_amino = "cosmos-sdk/MsgBeginRedelegate"
    """"""
    type_url = "/cosmos.staking.v1beta1.MsgBeginRedelegate"
    """"""
    action = "begin_redelegate"
    """"""
    prototype = MsgBeginRedelegate_pb
    """"""

    delegator_address: AccAddress = attr.ib()
    validator_src_address: ValAddress = attr.ib()
    validator_dst_address: ValAddress = attr.ib()
    amount: Coin = attr.ib(converter=Coin.parse)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_src_address": self.validator_src_address,
                "validator_dst_address": self.validator_dst_address,
                "amount": self.amount.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgBeginRedelegate:
        return cls(
            delegator_address=data["delegator_address"],
            validator_src_address=data["validator_src_address"],
            validator_dst_address=data["validator_dst_address"],
            amount=Coin.from_data(data["amount"]),
        )

    def to_proto(self) -> MsgBeginRedelegate_pb:
        return MsgBeginRedelegate_pb(
            delegator_address=self.delegator_address,
            validator_src_address=self.validator_src_address,
            validator_dst_address=self.validator_dst_address,
            amount=self.amount.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgBeginRedelegate_pb) -> MsgBeginRedelegate:
        return cls(
            delegator_address=proto.delegator_address,
            validator_src_address=proto.validator_src_address,
            validator_dst_address=proto.validator_dst_address,
            amount=Coin.from_proto(proto.amount),
        )


@attr.s
class MsgDelegate(Msg):
    """Delegate Luna to validator at ``validator_address``.

    Args:
        delegator_address: delegator
        validator_address: validator to delegate to
        amount (Union[str, dict, Coin]): coin (LUNA) to delegate
    """

    type_amino = "cosmos-sdk/MsgDelegate"
    """"""
    type_url = "/cosmos.staking.v1beta1.MsgDelegate"
    """"""
    action = "delegate"
    """"""
    prototype = MsgDelegate_pb
    """"""

    delegator_address: AccAddress = attr.ib()
    validator_address: ValAddress = attr.ib()
    amount: Coin = attr.ib(converter=Coin.parse)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "amount": self.amount.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgDelegate:
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            amount=Coin.from_data(data["amount"]),
        )

    def to_proto(self) -> MsgDelegate_pb:
        return MsgDelegate_pb(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            amount=self.amount.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgDelegate_pb) -> MsgDelegate:
        return cls(
            delegator_address=proto.delegator_address,
            validator_address=proto.validator_address,
            amount=Coin.from_proto(proto.amount),
        )


@attr.s
class MsgUndelegate(Msg):
    """Undelegate Luna from staking position with ``validator_address``.

    Args:
        delegator_address: delegator
        validator_address: validator to undelegate from
        amount (Union[str, dict, Coin]): coin (LUNA) to undelegate
    """

    type_amino = "cosmos-sdk/MsgUndelegate"
    """"""
    type_url = "/cosmos.staking.v1beta1.MsgUndelegate"
    """"""
    action = "begin_unbonding"
    """"""
    prototype = MsgUndelegate_pb
    """"""

    delegator_address: AccAddress = attr.ib()
    validator_address: ValAddress = attr.ib()
    amount: Coin = attr.ib(converter=Coin.parse)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "amount": self.amount.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgUndelegate:
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            amount=Coin.from_data(data["amount"]),
        )

    def to_proto(self) -> MsgUndelegate_pb:
        return MsgUndelegate_pb(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            amount=self.amount.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgUndelegate_pb) -> MsgUndelegate:
        return cls(
            delegator_address=proto.delegator_address,
            validator_address=proto.validator_address,
            amount=Coin.from_proto(proto.amount),
        )


@attr.s
class MsgEditValidator(Msg):
    """Revise validator description and configuration.

    Args:
        description: updated validator description
        validator_address: validator operator address
        commission_rates: new validator commission rate,
        min_self_delegation: new minimum self delegation,
    """

    type = "cosmos-sdk/MsgEditValidator"
    """"""
    type_url = "/cosmos.staking.v1beta1.MsgEditValidator"
    """"""
    action = "edit_validator"
    """"""
    prototype = MsgEditValidator_pb
    """"""

    description: Description = attr.ib()
    validator_address: ValAddress = attr.ib()
    commission_rate: Optional[Dec] = attr.ib(default=None)
    min_self_delegation: Optional[int] = attr.ib(default=None)

    @classmethod
    def from_data(cls, data: dict) -> MsgEditValidator:
        msd = int(data["min_self_delegation"]) if data["min_self_delegation"] else None
        cr = Dec(data["commission_rate"]) if data["commission_rate"] else None
        return cls(
            description=data["description"],
            validator_address=data["validator_address"],
            commission_rate=cr,
            min_self_delegation=msd,
        )

    def to_proto(self) -> MsgEditValidator_pb:
        return MsgEditValidator_pb(
            description=self.description.to_proto(),
            validator_address=self.validator_address,
            commission_rate=str(self.commission_rate) if self.commission_rate else None,
            min_self_delegation=str(self.min_self_delegation)
            if self.min_self_delegation
            else None,
        )

    @classmethod
    def from_proto(cls, proto: MsgEditValidator_pb) -> MsgEditValidator:
        msd = int(proto.min_self_delegation) if proto.min_self_delegation else "0"
        cr = Dec(proto.commission_rate) if proto.commission_rate else Dec("0")
        return cls(
            description=proto.description,
            validator_address=proto.validator_address,
            commission_rate=cr,
            min_self_delegation=msd,
        )


@attr.s
class MsgCreateValidator(Msg):
    """Register a new validator with the Terra protocol.

    Args:
        description: validator description
        commission: validator commission rates
        min_self_delegation: minimum self-delegation policy
        delegator_address: validator's account address
        validator_address: validator's operator address
        pubkey: validator consensus (Tendermint) public key
        value (Coin.Input): initial amount of Luna toi self-delegate
    """

    type = "cosmos-sdk/MsgCreateValidator"
    """"""
    type_url = "/cosmos.staking.v1beta1.MsgCreateValidator"
    """"""
    action = "create_validator"
    """"""
    prototype = MsgCreateValidator_pb
    """"""

    description: Description = attr.ib()
    commission: CommissionRates = attr.ib()
    min_self_delegation: str = attr.ib()
    delegator_address: AccAddress = attr.ib()
    validator_address: ValAddress = attr.ib()
    pubkey: ValConsPubKey = attr.ib()
    value: Coin = attr.ib(converter=Coin.parse)  # type: ignore

    @classmethod
    def from_data(cls, data: dict) -> MsgCreateValidator:
        return cls(
            description=Description.from_data(data["description"]),
            commission=CommissionRates.from_data(data["commission"]),
            min_self_delegation=int(data["min_self_delegation"]),
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            pubkey=data["pubkey"],
            value=Coin.from_data(data["value"]),
        )

    def to_proto(self) -> MsgCreateValidator_pb:
        return MsgCreateValidator_pb(
            description=self.description.to_proto(),
            commission=self.commission.to_proto(),
            min_self_delegation=self.min_self_delegation,
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            pubkey=self.pubkey.pack_any(),
            value=self.value.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgCreateValidator_pb) -> MsgCreateValidator:
        return cls(
            description=Description.from_proto(proto.description),
            commission=CommissionRates.from_proto(proto.commission),
            min_self_delegation=proto.min_self_delegation,
            delegator_address=proto.delegator_address,
            validator_address=proto.validator_address,
            pubkey=PublicKey.unpack_any(proto.pubkey),
            value=Coin.from_proto(proto.value),
        )



"""Params module governance proposal types."""

from __future__ import annotations

from typing import List

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.params.v1beta1 import ParamChange as ParamChange_pb
from terra_proto.cosmos.params.v1beta1 import (
    ParameterChangeProposal as ParameterChangeProposal_pb,
)

from terra_sdk.util.json import JSONSerializable

__all__ = ["ParameterChangeProposal", "ParamChange"]


@attr.s
class ParamChange(JSONSerializable):
    subspace: str = attr.ib()
    key: str = attr.ib()
    value: str = attr.ib()

    def to_amino(self) -> dict:
        return {"subspace": self.subspace, "key": self.key, "value": self.value}

    @classmethod
    def from_data(cls, data: dict) -> ParamChange:
        return cls(subspace=data["subspace"], key=data["key"], value=data["value"])

    def to_proto(self) -> ParamChange_pb:
        return ParamChange_pb(subspace=self.subspace, key=self.key, value=self.value)

    def to_data(self) -> dict:
        return {"subspace": self.subspace, "key": self.key, "value": self.value}

    @classmethod
    def from_proto(cls, proto: ParamChange_pb) -> ParamChange:
        return cls(subspace=proto.subspace, key=proto.key, value=proto.value)


@attr.s
class ParameterChangeProposal(JSONSerializable):
    """Proposal to alter the blockchain parameters. Changes would be effective
    as soon as the proposal is passed.

    Args:
        title: proposal title
        description: proposal description
        change (List[ParamChange]): list of parameter changes
    """

    type_amino = "cosmos-sdk/ParameterChangeProposal"
    """"""
    type_url = "/cosmos.params.v1beta1.ParameterChangeProposal"
    """"""
    prototype = ParameterChangeProposal_pb
    """"""

    title: str = attr.ib()
    description: str = attr.ib()
    changes: List[ParamChange] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "changes": [change.to_amino() for change in self.changes],
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> ParameterChangeProposal:
        return cls(
            title=data["title"],
            description=data["description"],
            changes=[ParamChange.from_data(change) for change in data["changes"]],
        )

    def to_proto(self) -> ParameterChangeProposal_pb:
        return ParameterChangeProposal_pb(
            title=self.title,
            description=self.description,
            changes=[change.to_proto() for change in self.changes],
        )

    def to_data(self) -> dict:
        return {
            "title": self.title,
            "description": self.description,
            "changes": [change.to_data() for change in self.changes],
        }

    @classmethod
    def from_proto(cls, proto: ParameterChangeProposal_pb) -> ParameterChangeProposal:
        return cls(
            title=proto.title,
            description=proto.description,
            changes=[ParamChange.from_proto(change) for change in proto.changes],
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))



"""Data objects pertaining to accounts."""

from __future__ import annotations

from itertools import starmap
from typing import List

import attr
from terra_proto.cosmos.vesting.v1beta1 import (
    DelayedVestingAccount as DelayedVestingAccount_pb,
)

from terra_sdk.core import Dec
from terra_sdk.util.json import JSONSerializable

from ...public_key import PublicKey
from .base_account import BaseAccount
from .base_vesting_account import BaseVestingAccount

__all__ = ["DelayedVestingAccount"]


@attr.s
class DelayedVestingAccount:
    """Stores information about an account with delayed vesting."""

    base_vesting_account: BaseVestingAccount = attr.ib()

    type_amino = "cosmos-sdk/DelayedVestingAccount"
    type_url = "/cosmos.vesting.v1beta1.DelayedVestingAccount"

    def get_sequence(self) -> int:
        return self.base_vesting_account.get_sequence()

    def get_account_number(self) -> int:
        return self.base_vesting_account.get_account_number()

    def get_public_key(self) -> PublicKey:
        return self.base_vesting_account.get_public_key()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "base_vesting_account": self.base_vesting_account.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "base_vesting_account": self.base_vesting_account.to_data(),
        }

    def to_proto(self) -> DelayedVestingAccount_pb:
        return DelayedVestingAccount_pb(
            base_vesting_account=self.base_vesting_account.to_proto(),
        )

    @classmethod
    def from_amino(cls, amino: dict) -> DelayedVestingAccount:
        amino = amino["value"]
        return cls(
            base_vesting_account=BaseVestingAccount.from_amino(
                {
                    "type": BaseVestingAccount.type_amino,
                    "value": amino["base_vesting_account"],
                }
            )
        )

    @classmethod
    def from_data(cls, data: dict) -> DelayedVestingAccount:
        return cls(
            base_vesting_account=BaseVestingAccount.from_data(
                data["base_vesting_account"]
            )
        )

    @classmethod
    def from_proto(cls, proto: DelayedVestingAccount_pb) -> DelayedVestingAccount:
        return cls(
            base_vesting_account=BaseVestingAccount.from_proto(
                proto.base_vesting_account
            )
        )



"""ibc commitment module data objects."""
from __future__ import annotations

import base64

import attr
from terra_proto.ibc.core.commitment.v1 import MerklePrefix as MerklePrefix_pb
from terra_proto.ibc.core.commitment.v1 import MerkleRoot as MerkleRoot_pb

from terra_sdk.util.json import JSONSerializable

__all__ = ["MerkleRoot", "MerklePrefix"]


@attr.s
class MerkleRoot(JSONSerializable):
    hash: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MerkleRoot:
        return cls(hash=(data["hash"]))

    def to_proto(self) -> MerkleRoot_pb:
        return MerkleRoot_pb(hash=base64.b64decode(self.hash))

    @classmethod
    def from_proto(cls, proto: MerkleRoot_pb) -> MerkleRoot:
        return cls(hash=base64.b64encode(proto.hash).decode())


@attr.s
class MerklePrefix(JSONSerializable):
    key_prefix: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MerklePrefix:
        return cls(key_prefix=data["key_prefix"])

    def to_proto(self) -> MerklePrefix_pb:
        return MerklePrefix_pb(key_prefix=base64.b64decode(self.key_prefix))

    @classmethod
    def from_proto(cls, proto: MerklePrefix_pb) -> MerklePrefix:
        return cls(key_prefix=base64.b64encode(proto.key_prefix).decode())



from .data import (
    Account,
    BaseAccount,
    ContinuousVestingAccount,
    DelayedVestingAccount,
    PeriodicVestingAccount,
    PublicKey,
    TxInfo,
    TxLog,
    parse_tx_logs,
)

__all__ = [
    "Account",
    "BaseAccount",
    "ContinuousVestingAccount",
    "DelayedVestingAccount",
    "PeriodicVestingAccount",
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
]
